# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import itertools
import logging

from gitfourchette.forms.stashdialog import StashDialog
from gitfourchette.gitdriver import GitStatus
from gitfourchette.localization import *
from gitfourchette.nav import NavLocator
from gitfourchette.porcelain import *
from gitfourchette.qt import *
from gitfourchette.tasks.repotask import AbortTask, RepoTask, TaskEffects, TaskPrereqs
from gitfourchette.toolbox import *
from gitfourchette.trash import Trash

logger = logging.getLogger(__name__)

_indexStatusTable = {
    GitStatus.Added: FileStatus.INDEX_NEW,
    GitStatus.Deleted: FileStatus.INDEX_DELETED,
    GitStatus.Modified: FileStatus.INDEX_MODIFIED,
    GitStatus.Renamed: FileStatus.INDEX_RENAMED,
    GitStatus.TypeChanged: FileStatus.INDEX_TYPECHANGE,
}
"""
Vanilla git staged file status to libgit2 index status flags
"""

_wtStatusTable = {
    GitStatus.Untracked: FileStatus.WT_NEW,
    GitStatus.Added: FileStatus.WT_NEW,
    GitStatus.Deleted: FileStatus.WT_DELETED,
    GitStatus.Modified: FileStatus.WT_MODIFIED,
    GitStatus.Renamed: FileStatus.WT_RENAMED,
    GitStatus.TypeChanged: FileStatus.WT_TYPECHANGE,
}
"""
Vanilla git unstaged file status to libgit2 worktree status flags
"""

_statusZero = FileStatus.CURRENT
assert _statusZero == 0


def backupStash(repo: Repo, stashCommitId: Oid):
    try:
        trashFile = Trash.instance().newFile(repo.workdir, ext=".txt", originalPath="DELETED_STASH")
    except Trash.BackupSkipped as ex:
        logger.warning(f"Stash backup skipped: {ex}")
        return

    text = F"""\
To recover this stash, paste the hash below into "Repo > Recall Lost Commit" in {qAppName()}:

{stashCommitId}

----------------------------------------

Original stash message below:

{repo.peel_commit(stashCommitId).message}
"""

    with open(trashFile, "w", encoding="utf-8") as f:
        f.write(text)


class NewStash(RepoTask):
    def prereqs(self):
        # libgit2 will refuse to create a stash if there are conflicts (NoConflicts)
        # libgit2 will refuse to create a stash if there are no commits at all (NoUnborn)
        return TaskPrereqs.NoConflicts | TaskPrereqs.NoUnborn

    def flow(self, paths: list[str] | None = None):
        gitStatus = yield from self.flowCallGit(
            "--no-optional-locks",
            "status",
            "--porcelain=v2",
            "-z",
            "--untracked-files=all")
        # Get GitDelta lists from 'git status' output
        numEntries, stagedDeltas, unstagedDeltas = gitStatus.readStatusPorcelainV2Z(self.repo.head_commit_id)

        if not numEntries:
            raise AbortTask(_("There are no uncommitted changes to stash."), "information")

        # Filter out any submodules or trees
        stagedDeltas   = [d for d in stagedDeltas   if not d.isTreeOrSubmodule()]
        unstagedDeltas = [d for d in unstagedDeltas if not d.isTreeOrSubmodule()]

        # If the selection only contained submodules, bail here
        if not stagedDeltas and not unstagedDeltas:
            raise AbortTask(_("Cannot stash submodules or subtrees."), "information")

        # Convert to libgit2 status flags for StashDialog
        status: dict[str, FileStatus] = {}
        for delta, statusConversion in itertools.chain(
                ((sd, _indexStatusTable) for sd in stagedDeltas),
                ((ud, _wtStatusTable) for ud in unstagedDeltas)
        ):
            path = delta.new.path
            bits = status.get(path, _statusZero)
            bits |= statusConversion.get(delta.status, _statusZero)
            status[path] = bits

        # Ask user what to stash
        dlg = StashDialog(status, paths or [], self.parentWidget())
        dlg.setWindowModality(Qt.WindowModality.WindowModal)
        dlg.show()
        yield from self.flowDialog(dlg)

        paths = dlg.tickedPaths()
        stashMessage = dlg.ui.messageEdit.text()
        keepIntact = dlg.ui.keepCheckBox.isChecked()
        dlg.deleteLater()

        # Rationale for sticking with libgit2 here: 'git stash' always brings in
        # ALL the staged changes, even if we explicitly pass some paths.
        yield from self.flowEnterWorkerThread()
        self.epilog.effects |= TaskEffects.Refs
        self.repo.create_stash(stashMessage, paths=paths)
        self.epilog.status = _n("File stashed.", "{n} files stashed.", len(paths))
        yield from self.flowEnterUiThread()

        if keepIntact:
            return

        # Clean up with vanilla git so smudge filters apply properly
        self.epilog.effects |= TaskEffects.Workdir

        # Clean untracked files
        cleanPaths = [p for p in paths if status[p] == FileStatus.WT_NEW]
        if cleanPaths:
            yield from self.flowCallGit("clean", "--force", "--", *cleanPaths)

        # Restore other files
        restorePaths = [p for p in paths if p not in cleanPaths]
        if restorePaths:
            yield from self.flowCallGit("restore", "--progress", "--source=HEAD", "--worktree", "--staged", "--", *restorePaths)


class ApplyStash(RepoTask):
    def prereqs(self):
        # libgit2 will refuse to apply a stash if there are conflicts (NoConflicts)
        return TaskPrereqs.NoConflicts | TaskPrereqs.NoStagedChanges

    def flow(self, stashCommitId: Oid, tickDelete=True):
        stashCommit: Commit = self.repo.peel_commit(stashCommitId)
        stashMessage = strip_stash_message(stashCommit.message)

        question = _("Do you want to apply the changes stashed in {0} "
                     "to your working directory?", bquoe(stashMessage))

        qmb = asyncMessageBox(self.parentWidget(), 'question', self.name(), question,
                              QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel,
                              deleteOnClose=False)

        def updateButtonText(ticked: bool):
            okButton = qmb.button(QMessageBox.StandardButton.Ok)
            okButton.setText(_("&Apply && Delete") if ticked else _("&Apply && Keep"))

        deleteCheckBox = QCheckBox(_("&Delete the stash if it applies cleanly"), qmb)
        deleteCheckBox.clicked.connect(updateButtonText)
        deleteCheckBox.setChecked(tickDelete)
        qmb.setCheckBox(deleteCheckBox)
        updateButtonText(tickDelete)
        yield from self.flowDialog(qmb)

        deleteAfterApply = deleteCheckBox.isChecked()
        qmb.deleteLater()

        self.epilog.jumpTo = NavLocator.inWorkdir()
        self.epilog.effects |= TaskEffects.Workdir

        if deleteAfterApply:
            self.epilog.effects |= TaskEffects.Refs
            backupStash(self.repo, stashCommitId)

        # Although 'git stash apply stash@{<FULL_COMMIT_ID>}' works fine,
        # 'git stash pop' and 'drop' won't delete the stash if we pass the full
        # commit id. So, use a stash number instead.
        stashIndex = self.repo.find_stash_index(stashCommitId)
        popOrApply = "pop" if deleteAfterApply else "apply"

        driver = yield from self.flowCallGit("stash", popOrApply, str(stashIndex), autoFail=False)

        self.repo.refresh_index()
        anyConflicts = self.repo.index.conflicts

        if driver.exitCode() == 0:
            if deleteAfterApply:
                self.epilog.status = _("Stash {0} applied and deleted.", tquoe(stashMessage))
            else:
                self.epilog.status = _("Stash {0} applied.", tquoe(stashMessage))

        elif anyConflicts:
            self.epilog.status = _("Stash {0} applied, with conflicts.", tquoe(stashMessage))
            message = [_("Applying the stash {0} has caused merge conflicts "
                         "because your files have diverged since they were stashed.", bquoe(stashMessage))]
            if deleteAfterApply:
                message.append(_("The stash wasn’t deleted in case you need to re-apply it later."))
            showWarning(self.parentWidget(), _("Conflicts caused by stash application"), paragraphs(*message))

        else:
            self.epilog.status = _("Stash {0} couldn’t be applied.", tquoe(stashMessage))
            message = [self.epilog.status]
            if deleteAfterApply:
                message.append(_("The stash wasn’t deleted in case you need to re-apply it later."))
            markup = driver.htmlErrorText(paragraphs(*message))
            raise AbortTask(markup, details=driver.formatCommandLine())


class DropStash(RepoTask):
    def flow(self, stashCommitId: Oid):
        stashCommit = self.repo.peel_commit(stashCommitId)
        stashMessage = strip_stash_message(stashCommit.message)
        yield from self.flowConfirm(
            text=_("Really delete stash {0}?", bquoe(stashMessage)),
            verb=_("Delete stash"),
            buttonIcon="SP_DialogDiscardButton")

        backupStash(self.repo, stashCommitId)

        self.epilog.effects |= TaskEffects.Refs
        stashIndex = self.repo.find_stash_index(stashCommitId)
        stashName = f"stash@{{{stashIndex}}}"  # 'git stash drop' doesn't support stash numbers
        yield from self.flowCallGit("stash", "drop", stashName)

        self.epilog.status = _("Stash {0} deleted.", tquoe(stashMessage))



# Form implementation generated from reading ui file 'reposettingsdialog.ui'
#
# Created by: PyQt6 UI code generator 6.10.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_RepoSettingsDialog(object):
    def setupUi(self, RepoSettingsDialog):
        RepoSettingsDialog.setObjectName("RepoSettingsDialog")
        RepoSettingsDialog.resize(465, 229)
        self.formLayout = QFormLayout(RepoSettingsDialog)
        self.formLayout.setFieldGrowthPolicy(QFormLayout.FieldGrowthPolicy.ExpandingFieldsGrow)
        self.formLayout.setObjectName("formLayout")
        self.nicknameLabel = QLabel(parent=RepoSettingsDialog)
        self.nicknameLabel.setObjectName("nicknameLabel")
        self.formLayout.setWidget(0, QFormLayout.ItemRole.LabelRole, self.nicknameLabel)
        self.nicknameEdit = QLineEdit(parent=RepoSettingsDialog)
        self.nicknameEdit.setClearButtonEnabled(True)
        self.nicknameEdit.setObjectName("nicknameEdit")
        self.formLayout.setWidget(0, QFormLayout.ItemRole.FieldRole, self.nicknameEdit)
        spacerItem = QSpacerItem(20, 8, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        self.formLayout.setItem(1, QFormLayout.ItemRole.LabelRole, spacerItem)
        self.localIdentityCheckBox = QCheckBox(parent=RepoSettingsDialog)
        self.localIdentityCheckBox.setObjectName("localIdentityCheckBox")
        self.formLayout.setWidget(2, QFormLayout.ItemRole.SpanningRole, self.localIdentityCheckBox)
        self.nameLabel = QLabel(parent=RepoSettingsDialog)
        self.nameLabel.setObjectName("nameLabel")
        self.formLayout.setWidget(3, QFormLayout.ItemRole.LabelRole, self.nameLabel)
        self.nameEdit = QLineEdit(parent=RepoSettingsDialog)
        self.nameEdit.setObjectName("nameEdit")
        self.formLayout.setWidget(3, QFormLayout.ItemRole.FieldRole, self.nameEdit)
        self.emailLabel = QLabel(parent=RepoSettingsDialog)
        self.emailLabel.setObjectName("emailLabel")
        self.formLayout.setWidget(4, QFormLayout.ItemRole.LabelRole, self.emailLabel)
        self.emailEdit = QLineEdit(parent=RepoSettingsDialog)
        self.emailEdit.setObjectName("emailEdit")
        self.formLayout.setWidget(4, QFormLayout.ItemRole.FieldRole, self.emailEdit)
        spacerItem1 = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.formLayout.setItem(7, QFormLayout.ItemRole.LabelRole, spacerItem1)
        self.buttonBox = QDialogButtonBox(parent=RepoSettingsDialog)
        self.buttonBox.setOrientation(Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Cancel|QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.formLayout.setWidget(8, QFormLayout.ItemRole.SpanningRole, self.buttonBox)
        self.keyFilePicker = KeyFilePickerCheckBox(parent=RepoSettingsDialog)
        self.keyFilePicker.setObjectName("keyFilePicker")
        self.formLayout.setWidget(6, QFormLayout.ItemRole.SpanningRole, self.keyFilePicker)
        spacerItem2 = QSpacerItem(20, 8, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        self.formLayout.setItem(5, QFormLayout.ItemRole.LabelRole, spacerItem2)
        self.nicknameLabel.setBuddy(self.nicknameEdit)
        self.nameLabel.setBuddy(self.nameEdit)
        self.emailLabel.setBuddy(self.emailEdit)

        self.retranslateUi(RepoSettingsDialog)
        self.buttonBox.accepted.connect(RepoSettingsDialog.accept)
        self.buttonBox.rejected.connect(RepoSettingsDialog.reject)
        self.localIdentityCheckBox.toggled['bool'].connect(self.nameLabel.setEnabled)
        self.localIdentityCheckBox.toggled['bool'].connect(self.nameEdit.setEnabled)
        self.localIdentityCheckBox.toggled['bool'].connect(self.emailLabel.setEnabled)
        self.localIdentityCheckBox.toggled['bool'].connect(self.emailEdit.setEnabled)

    def retranslateUi(self, RepoSettingsDialog):
        RepoSettingsDialog.setWindowTitle(_p("RepoSettingsDialog", "Repo Settings for {repo}"))
        self.nicknameLabel.setText(_p("RepoSettingsDialog", "Repo Nic&kname:"))
        self.nicknameEdit.setToolTip(_p("RepoSettingsDialog", "This nickname will appear within {app} in tab names, menus, etc. It does not change the actual name of the repo’s directory. Leave blank to clear the nickname."))
        self.nicknameEdit.setPlaceholderText(_p("RepoSettingsDialog", "No nickname"))
        self.localIdentityCheckBox.setText(_p("RepoSettingsDialog", "Create commits under a custom &identity in this repo"))
        self.nameLabel.setText(_p("RepoSettingsDialog", "&Name:"))
        self.emailLabel.setText(_p("RepoSettingsDialog", "&Email:"))
        self.keyFilePicker.setText(_p("RepoSettingsDialog", "Log in to SSH remotes with custom &key file"))
from gitfourchette.forms.keyfilepickercheckbox import KeyFilePickerCheckBox



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import logging
import re
import shlex
from contextlib import suppress

from gitfourchette.exttools.toolcommands import ToolCommands
from gitfourchette.forms.ui_statusform import Ui_StatusForm
from gitfourchette.gitdriver import GitDriver
from gitfourchette.localization import _
from gitfourchette.qt import *
from gitfourchette.toolbox import stockIcon, tweakWidgetFont, QProcessConnection

logger = logging.getLogger(__name__)

# Extract the verb part from a git command, e.g. "cherry-pick" when the command
# is "git cherry-pick". See testGitVerbPattern for examples of inputs.
_gitVerbPattern = re.compile(
    r"git(?:\.exe|\.EXE)?['\"]?(?=\s).*?\s"
    r"((?:(lfs|submodule)\s+)?[a-z][a-z\-]*)"
    r"(?=$|\s)")


class StatusForm(QWidget):
    def __init__(self, parent):
        super().__init__(parent)

        self.processConnection = QProcessConnection(self)
        self.processConnection.processLost.connect(self.onProcessLost)

        self.sentSigterm = False
        self.abortButton = None
        self.abortButtonDefaultText = ""
        self.abortButtonDefaultIcon = None

        self.ui = Ui_StatusForm()
        self.ui.setupUi(self)

        # Fix slightly different background color in blurb page in Fusion and macOS styles
        self.ui.blurbScrollArea.viewport().setAutoFillBackground(False)
        self.ui.blurbScrollAreaWidgetContents.setAutoFillBackground(False)

        tweakWidgetFont(self.ui.titleLabel, bold=True)
        tweakWidgetFont(self.ui.commandLabel, relativeSize=88)
        tweakWidgetFont(self.ui.statusLabel, tabularNumbers=True)

    def connectAbortButton(self, abortButton: QPushButton):
        self.abortButton = abortButton
        self.abortButtonDefaultText = abortButton.text()
        self.abortButtonDefaultIcon = abortButton.icon()

    def setBlurb(self, text: str):
        self.ui.stackedWidget.setCurrentIndex(0)
        self.ui.blurbLabel.setText(text)

    def initProgress(self, text: str):
        self.ui.stackedWidget.setCurrentIndex(1)
        self.setProgressMessage(text)
        self.ui.progressBar.setMinimum(0)
        self.setProgressValue(0, 0)

    def setProgressValue(self, value: int, maximum: int):
        self.ui.progressBar.setValue(value)
        self.ui.progressBar.setMaximum(maximum)

    def setProgressMessage(self, message: str):
        self.ui.statusLabel.setText(message)

    def connectProcess(self, process: QProcess):
        self.processConnection.track(process)
        self.sentSigterm = False

        commandLine = shlex.join([process.program()] + process.arguments())
        self.ui.commandLabel.setText(commandLine)

        gitVerbMatch = _gitVerbPattern.search(commandLine)
        if gitVerbMatch:
            title = f"git {gitVerbMatch.group(1)}…"
        else:
            title = f"{process.program()}…"

        self.ui.titleLabel.setText(title)
        self.initProgress(_("Please wait…"))
        self.setProgressValue(0, 0)

        assert self.abortButton is not None, "abort button not connected"
        self.abortButton.setText(_("Abort"))
        self.abortButton.setIcon(stockIcon("SP_DialogCloseButton"))
        self.abortButton.clearFocus()

        if isinstance(process, GitDriver):
            process.progressMessage.connect(self.setProgressMessage)
            process.progressFraction.connect(self.setProgressValue)

    def onProcessLost(self):
        process = self.processConnection.process
        assert process is not None
        if isinstance(process, GitDriver):
            with suppress(TypeError, RuntimeError):
                process.progressMessage.disconnect(self.setProgressMessage)
            with suppress(TypeError, RuntimeError):
                process.progressFraction.disconnect(self.setProgressValue)

        assert self.abortButton is not None, "abort button not connected"
        self.abortButton.setText(self.abortButtonDefaultText)
        self.abortButton.setIcon(self.abortButtonDefaultIcon)

    def requestAbort(self):
        self.setProgressMessage(_("Aborting…"))
        self.setProgressValue(0, 0)

        if self.abortButton is not None and not self.sentSigterm:
            self.abortButton.setText("SIGKILL")
            self.abortButton.setIcon(stockIcon("sigkill"))

        process = self.processConnection.process
        if not process:
            return
        elif not self.sentSigterm:
            ToolCommands.terminatePlus(process)
            self.sentSigterm = True
            self.ui.titleLabel.setText(_("SIGTERM sent. Waiting for process to terminate…"))
        else:
            process.kill()



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import os

from gitfourchette import settings
from gitfourchette.exttools.toolprocess import ToolProcess
from gitfourchette.filelists.filelist import FileList
from gitfourchette.gitdriver import GitDelta
from gitfourchette.localization import *
from gitfourchette.nav import NavLocator, NavContext
from gitfourchette.porcelain import *
from gitfourchette.qt import *
from gitfourchette.repomodel import RepoModel
from gitfourchette.settings import getExternalEditorName
from gitfourchette.tasks import *
from gitfourchette.toolbox import *


class CommittedFiles(FileList):
    def __init__(self, repoModel: RepoModel, parent: QWidget):
        super().__init__(repoModel, parent, NavContext.COMMITTED)

    def contextMenuActions(self, deltas: list[GitDelta]) -> list[ActionDef]:
        actions = []

        n = len(deltas)
        modeSet = {delta.new.mode for delta in deltas}  # modeDst
        anySubmodules = FileMode.COMMIT in modeSet
        onlySubmodules = anySubmodules and len(modeSet) == 1

        if not anySubmodules:
            actions += [
                self.contextMenuActionBlame(deltas),

                ActionDef.SEPARATOR,

                ActionDef(
                    _("Open Diff in New &Window"),
                    self.wantOpenDiffInNewWindow,
                ),

                *self.contextMenuActionsDiff(deltas),

                ActionDef.SEPARATOR,

                ActionDef(
                    _n("&Revert This Change…", "&Revert These Changes…", n),
                    self.revertPaths,
                ),

                ActionDef(
                    _("Restor&e File Revision…"),
                    submenu=[
                        ActionDef(_("&As Of This Commit"), self.restoreNewRevision),
                        ActionDef(_("&Before This Commit"), self.restoreOldRevision),
                    ]
                ),

                ActionDef.SEPARATOR,

                ActionDef(
                    _n("&Open File in {0}", "&Open {n} Files in {0}", n, settings.getExternalEditorName()),
                    icon="SP_FileIcon", submenu=[
                        ActionDef(_("&Current Revision (Working Copy)"), self.openWorkingCopyRevision),
                        ActionDef(_("&As Of This Commit"), self.openNewRevision),
                        ActionDef(_("&Before This Commit"), self.openOldRevision),
                    ]
                ),

                ActionDef(
                    _("&Save a Copy…"),
                    icon="SP_DialogSaveButton", submenu=[
                        ActionDef(_("&As Of This Commit"), self.saveNewRevision),
                        ActionDef(_("&Before This Commit"), self.saveOldRevision),
                    ]
                ),
            ]

        elif onlySubmodules:
            actions += [
                ActionDef(
                    _n("Submodule", "{n} Submodules", n),
                    kind=ActionDef.Kind.Section,
                ),

                ActionDef(
                    _n("Open Submodule in New Tab", "Open {n} Submodules in New Tabs", n),
                    self.openSubmoduleTabs,
                ),
            ]

        else:
            sorry = _("Please review the files individually.")
            actions += [
                ActionDef(sorry, enabled=False),
            ]

        actions += super().contextMenuActions(deltas)
        return actions

    def setCommitLocator(self, locator: NavLocator):
        assert locator.context == NavContext.COMMITTED
        self.flModel.navLocator = locator.coarse().replace(path="")

    def openNewRevision(self):
        self.openRevision(beforeCommit=False)

    def openOldRevision(self):
        self.openRevision(beforeCommit=True)

    def saveNewRevision(self):
        self.saveRevisionAs(beforeCommit=False)

    def saveOldRevision(self):
        self.saveRevisionAs(beforeCommit=True)

    def _restoreRevision(self, old: bool):
        deltas = list(self.selectedDeltas())
        assert len(deltas) == 1
        RestoreRevisionToWorkdir.invoke(self, deltas[0], old=old)

    def restoreNewRevision(self):
        self._restoreRevision(old=False)

    def restoreOldRevision(self):
        self._restoreRevision(old=True)

    # TODO: Send all files to text editor in one command?
    def openRevision(self, beforeCommit: bool = False):
        def run(task: RepoTask, delta: GitDelta):
            yield from task.flowSubtask(OpenRevisionInEditor, delta, beforeCommit)

        toolName = getExternalEditorName()
        self.confirmBatch(
            run,
            _("Open file revision"),
            _("Really open [# files] in {0}?", toolName))

    def saveRevisionAs(self, beforeCommit: bool = False):
        def run(task: RepoTask, delta: GitDelta):
            yield from task.flowSubtask(SaveRevisionAs, delta, old=beforeCommit)

        self.confirmBatch(run, _("Save file revision as"), _("Really export [# files]?"))

    def openWorkingCopyRevision(self):
        def run(task: RepoTask, delta: GitDelta):
            path = task.repo.in_workdir(delta.new.path)
            if not os.path.isfile(path):
                raise FileNotFoundError(_("There’s no file at this path in the working copy."))
            ToolProcess.startTextEditor(task.parentWidget(), path)
            yield from task.flowEnterUiThread()  # dummy yield

        toolName = getExternalEditorName()
        self.confirmBatch(
            run,
            _("Open working copy revision"),
            _("Really open [# files] in {0}?", toolName))

    def wantOpenDiffInNewWindow(self):
        sourceLocator = self.flModel.navLocator

        def run(task: RepoTask, delta: GitDelta):
            locator = sourceLocator.replace(path=delta.new.path)
            yield from task.flowSubtask(LoadPatchInNewWindow, delta, locator)

        self.confirmBatch(run, _("Open diff in new window"), _("Really open [# windows]?"))



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import re
from contextlib import suppress
from typing import Literal

from gitfourchette import trtables
from gitfourchette.porcelain import *
from gitfourchette.qt import *

INITIALS_PATTERN = re.compile(r"(?:^|[\s\-.'‘’\"“”])+([^\s\-.'‘’\"“”])[^\s\-.]*")
FIRST_NAME_PATTERN = re.compile(r"(\S(\.?-|\.\s?|\s))*\S+")
WINDOWS_PATH_PATTERN = re.compile(r"['\"]?^[A-Za-z]:[/\\]")

REMOTE_URL_PATTERNS = [
    # HTTP/HTTPS
    # http://example.com/user/repo
    # https://example.com/user/repo
    # https://personal_access_token@example.com/user/repo (GitHub write access over HTTPS)
    re.compile(r"^(?P<protocol>https?):\/\/(?P<user>[^@/]+@)?(?P<host>[^\/]+?)\/(?P<path>.+)"),

    # SSH (scp-like syntax)
    # example.com:user/repo
    # git@example.com:user/repo
    re.compile(r"^(?P<user>[^@/]+@)?(?P<host>[^\/]+?):(?!\/)(?P<path>.+)"),

    # SSH (full syntax)
    # ssh://example.com/user/repo
    # ssh://git@example.com/user/repo
    # ssh://git@example.com:1234/user/repo
    re.compile(r"^(?P<protocol>ssh):\/\/(?P<user>[^@/]+@)?(?P<host>[^\/]+?)(:\d+)?\/(?P<path>.+)"),

    # Git protocol
    # git://example.com/user/repo
    # git://example.com:1234/user/repo
    re.compile(r"^(?P<protocol>git):\/\/(?P<host>[^\/]+?)(:\d+)?\/(?P<path>.+)"),
]


class AuthorDisplayStyle(enum.IntEnum):
    FullName = 1
    FirstName = 2
    LastName = 3
    Initials = 4
    FullEmail = 5
    EmailUserName = 6


@enum.unique
class PatchPurpose(enum.IntFlag):
    Stage = enum.auto()
    Unstage = enum.auto()
    Discard = enum.auto()

    Lines = enum.auto()
    Hunk = enum.auto()
    File = enum.auto()

    VerbMask = Stage | Unstage | Discard


def abbreviatePerson(sig: Signature, style: AuthorDisplayStyle = AuthorDisplayStyle.FullName):
    with suppress(IndexError):
        if style == AuthorDisplayStyle.FullName:
            return sig.name

        elif style == AuthorDisplayStyle.FirstName:
            match = FIRST_NAME_PATTERN.match(sig.name)
            return match[0] if match is not None else sig.name

        elif style == AuthorDisplayStyle.LastName:
            return sig.name.rsplit(' ', maxsplit=1)[-1]

        elif style == AuthorDisplayStyle.Initials:
            return re.sub(INITIALS_PATTERN, r"\1", sig.name)

        elif style == AuthorDisplayStyle.FullEmail:
            return sig.email

        elif style == AuthorDisplayStyle.EmailUserName:
            emailParts = sig.email.split('@', 1)
            if len(emailParts) == 2 and emailParts[1] == "users.noreply.github.com":
                # Strip ID from GitHub noreply addresses (1234567+username@users.noreply.github.com)
                return emailParts[0].split('+', 1)[-1]
            else:
                return emailParts[0]

    return sig.email  # type: ignore[unreachable]


def shortHash(oid: Oid | str) -> str:
    from gitfourchette.settings import prefs
    return str(oid)[:prefs.shortHashChars]


def nameValidationMessage(name: str, reservedNames: list[str], nameTakenMessage: str = "") -> str:
    try:
        validate_refname(name, reservedNames)
    except NameValidationError as exc:
        if exc.rule == NameValidationError.Rule.NAME_TAKEN_BY_REF and nameTakenMessage:
            return nameTakenMessage
        else:
            return trtables.enum(exc.rule)

    return ""  # validation passed, no error


def remoteUrlProtocol(url: str):
    # Bail early on Windows-style absolute paths (C:\Whatever) to avoid looking like an ssh url
    if WINDOWS_PATH_PATTERN.match(url):
        return ""

    for pattern in REMOTE_URL_PATTERNS:
        m = pattern.match(url)
        if m:
            try:
                return m.group("protocol")
            except IndexError:
                return "ssh"
    return ""


def splitRemoteUrl(url: str):
    for pattern in REMOTE_URL_PATTERNS:
        m = pattern.match(url)
        if m:
            host = m.group("host")
            path = m.group("path")
            return host, path
    return "", ""


def guessRemoteUrlFromText(text: str):
    if len(text) > 128:
        return ""

    text = text.strip()

    if any(c.isspace() for c in text):
        return ""

    host, path = splitRemoteUrl(text)
    if host and path:
        return text

    return ""


def formatTimeOffset(minutes: int):
    p = "-" if minutes < 0 else "+"
    h = abs(minutes) // 60
    m = abs(minutes) % 60
    return f"{p}{h:02}:{m:02}"


def signatureEnvironmentVariables(sig: Signature, infix: Literal["AUTHOR", "COMMITTER"]) -> dict[str, str]:
    return {
        f"GIT_{infix}_NAME": sig.name,
        f"GIT_{infix}_EMAIL": sig.email,
        f"GIT_{infix}_DATE": f"{sig.time}{formatTimeOffset(sig.offset)}",
    }


def signatureQDateTime(signature: Signature, localTime=False) -> QDateTime:
    if localTime:
        return QDateTime.fromSecsSinceEpoch(signature.time)
    else:
        return QDateTime.fromSecsSinceEpoch(signature.time, QTimeZone(signature.offset * 60))


def signatureDateFormat(
        signature: Signature,
        format: str | QLocale.FormatType = QLocale.FormatType.LongFormat,
        localTime=False
) -> str:
    dateTime = signatureQDateTime(signature, localTime)
    text = QLocale().toString(dateTime, format)
    if not localTime and format != QLocale.FormatType.LongFormat:
        text += f" ({formatTimeOffset(signature.offset)})"
    return text



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import dataclasses
import enum

from pygit2.enums import AttrCheck

from gitfourchette.gitdriver.gitconflict import GitConflict
from gitfourchette.gitdriver.gitdeltafile import GitDeltaFile, FileMode, GitDeltaSource
from gitfourchette.gitdriver.lfspointer import LfsPointer, LfsPointerState
from gitfourchette.porcelain import Repo


class GitStatus(enum.StrEnum):
    """Status of a file in a GitDelta."""

    # git: diff.h "diff-raw status letters"
    Added = "A"
    Copied = "C"
    Deleted = "D"
    Modified = "M"
    Renamed = "R"
    TypeChanged = "T"
    Unmerged = "U"  # aka merge conflict
    Unknown = "X"  # aka Unreadable in libgit2

    # Additional statuses
    Untracked = "?"
    Ignored = "!"  # note: "I" in libgit2 (git_diff_status_char())

    @property
    def isAddedOrUntracked(self) -> bool:
        return self in "?A"


@dataclasses.dataclass
class GitDelta:
    status: GitStatus = GitStatus.Unknown
    old: GitDeltaFile = dataclasses.field(default_factory=GitDeltaFile)
    new: GitDeltaFile = dataclasses.field(default_factory=GitDeltaFile)
    similarity: int = 0
    submoduleStatus: str = ""  # Only in UNSTAGED contexts
    conflict: GitConflict | None = None  # Only in UNSTAGED contexts

    def __post_init__(self):
        assert isinstance(self.status, GitStatus)
        assert self.old.source != GitDeltaSource.Dirty, "old source cannot be dirty"

        # Clear empty submodule status so it's falsy
        if self.submoduleStatus == "N...":
            self.submoduleStatus = ""

    @property
    def source(self) -> GitDeltaSource:
        return self.new.source

    @property
    def submoduleWorkdirDirty(self) -> bool:
        sub = self.submoduleStatus
        return "M" in sub or "U" in sub

    def isSubtreeCommitPatch(self) -> bool:
        return (self.old.mode | self.new.mode) & FileMode.COMMIT == FileMode.COMMIT

    def isTreeOrSubmodule(self) -> bool:
        assert (FileMode.COMMIT & FileMode.TREE) == FileMode.TREE  # commit must contain tree bit
        return (self.old.mode | self.new.mode) & FileMode.TREE == FileMode.TREE

    def cacheLfsPointers(self, repo: Repo):
        old = self.old
        new = self.new

        # Cache "old" LFS pointer
        if old.lfs.state:
            # Already cached
            pass
        elif self.status.isAddedOrUntracked:
            # Untracked/unstaged: No pointer yet
            old.lfs = LfsPointer(LfsPointerState.NoPointer)
        else:
            if old.source == GitDeltaSource.Commit:
                oldCheck = AttrCheck.INCLUDE_COMMIT
            else:
                assert old.source != GitDeltaSource.Dirty, "old source cannot be dirty"
                oldCheck = AttrCheck.INDEX_THEN_FILE
                if not repo.head_is_unborn:
                    oldCheck |= AttrCheck.INCLUDE_HEAD

            old.cacheLfsPointer(repo, oldCheck)

        # Cache "new" LFS pointer
        if new.lfs.state:
            # Already cached
            pass
        elif self.status == GitStatus.Deleted:
            # Deletion: No pointer
            new.lfs = LfsPointer(LfsPointerState.NoPointer)
        else:
            if new.source == GitDeltaSource.Commit:
                newCheck = AttrCheck.INCLUDE_COMMIT
            elif new.source == GitDeltaSource.Dirty:
                # Note: If .gitattributes itself contains unstaged changes, then
                # this check is unreliable with libgit2 alone (we'd need an
                # AttrCheck.FILE_ONLY flag). For that specific case,
                # `loadWorkdir` should already have cached the 'new' lfs pointer
                # state via `git check-attr`.
                newCheck = AttrCheck.FILE_THEN_INDEX
            else:
                newCheck = AttrCheck.INDEX_ONLY

            new.cacheLfsPointer(repo, newCheck)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations

import dataclasses
import logging
import shlex
from contextlib import suppress
from pathlib import Path

from gitfourchette import settings
from gitfourchette.exttools.toolcommands import ToolCommands
from gitfourchette.exttools.toolprocess import ToolProcess
from gitfourchette.localization import *
from gitfourchette.porcelain import Oid
from gitfourchette.qt import *
from gitfourchette.toolbox import *

logger = logging.getLogger(__name__)

try:
    import gitfourchette.mount.treemount
    fuseImportError = None
except (ImportError, OSError) as exc:  # pragma: no cover
    fuseImportError = exc
    logger.info(f"FUSE not available: {fuseImportError}")


@dataclasses.dataclass
class MountedCommit:
    repoWorkdir: str
    commitId: Oid
    mountPoint: str
    fuseProcess: QProcess

    def friendlyName(self) -> str:
        repoName = settings.history.getRepoNickname(self.repoWorkdir)
        return f"{repoName} @ {shortHash(self.commitId)}"

    def openFolder(self):
        openFolder(self.mountPoint)

    def openTerminal(self, parent: QWidget):
        ToolProcess.startTerminal(parent, self.mountPoint)

    def unmount(self):
        ToolCommands.terminatePlus(self.fuseProcess)
        self.fuseProcess.waitForFinished(5000)

        if not WINDOWS:  # WinFSP fully manages the mount point folder
            with suppress(OSError):
                Path(self.mountPoint).rmdir()


class MountManager(QObject):
    mountPointsChanged = Signal()
    statusMessage = Signal(str)

    mountedCommits: dict[Oid, MountedCommit]

    def __init__(self, parent):
        super().__init__(parent)
        self.mountedCommits = {}

    @classmethod
    def isFuseAvailable(cls):
        return fuseImportError is None

    @classmethod
    def requireFuse(cls):
        if fuseImportError is not None:
            raise fuseImportError

    @classmethod
    def supportsMounting(cls) -> bool:
        return fuseImportError is None

    def isMounted(self, oid: Oid):
        return oid in self.mountedCommits

    def checkAliveProcesses(self):
        """
        Check if any processes have been terminated outside our control,
        e.g. by "ejecting" a drive on macOS.
        """
        alive = {oid: mc for oid, mc in self.mountedCommits.items()
                 if mc.fuseProcess.state() != QProcess.ProcessState.NotRunning}
        if len(alive) != len(self.mountedCommits):
            self.mountedCommits.clear()
            self.mountedCommits.update(alive)
            self.mountPointsChanged.emit()

    def mount(self, workdir: str, oid: Oid):
        self.requireFuse()

        if FLATPAK:
            # Flatpak version can't mount inside qTempDir, so create the mount
            # point somewhere on the disk. Avoid the repo's .git directory.
            mountParentDir = QStandardPaths.writableLocation(QStandardPaths.StandardLocation.GenericCacheLocation)
        else:
            mountParentDir = qTempDir()

        wdName = Path(workdir).name
        pathObj = Path(mountParentDir, f"mnt-{wdName}@{str(oid)[:8]}")
        if pathObj.is_dir():
            pathObj.rmdir()
        if not WINDOWS:  # WinFSP insists on creating it itself
            pathObj.mkdir(parents=True)
        path = str(pathObj)

        tokens = ToolCommands.spawnNewInstance(gitfourchette.mount.treemount.__name__, sandbox=True)
        tokens += [workdir, str(oid), path]
        logger.info(f"Starting: {shlex.join(tokens)}")

        fuseProcess = QProcess(self)
        fuseProcess.setProgram(tokens[0])
        fuseProcess.setArguments(tokens[1:])

        mc = MountedCommit(workdir, oid, path, fuseProcess)
        self.mountedCommits[oid] = mc

        fuseProcess.started.connect(lambda: self.onProcessStarted(mc))
        fuseProcess.start()
        logger.info(f"PID {fuseProcess.processId()}, mountpoint {path}")

        self.statusMessage.emit(
            _("Created a FUSE mount point for {0}. Find it in the {1} menu.",
              tquo(mc.friendlyName()),
              tquo(stripAccelerators(_("&Mount")))))

        return mc

    def onProcessStarted(self, mc: MountedCommit):
        self.mountPointsChanged.emit()
        mc.openFolder()

    def copyPath(self, mc: MountedCommit):
        QApplication.clipboard().setText(mc.mountPoint)
        self.statusMessage.emit(clipboardStatusMessage(mc.mountPoint))

    def unmount(self, mc: MountedCommit):
        mc.unmount()
        del self.mountedCommits[mc.commitId]
        self.statusMessage.emit(_("Unmounted {0}.", tquo(mc.friendlyName())))
        self.mountPointsChanged.emit()

    def unmountAll(self, closeCallback=None):
        for m in self.mountedCommits.values():
            m.unmount()
        self.mountedCommits.clear()
        self.statusMessage.emit(_("Unmounted all commit folders."))
        self.mountPointsChanged.emit()
        if closeCallback is not None:
            closeCallback()

    def makeMenu(self, widget: QWidget):
        items = []
        numMounts = len(self.mountedCommits)

        for oid, mc in self.mountedCommits.items():
            items.append(ActionDef(escamp(mc.friendlyName()), enabled=False))
            items.extend(self.makeMenuItemsForMount(oid, widget, numMounts == 1))
            items.append(ActionDef.SEPARATOR)

        if numMounts >= 2:
            items.append(ActionDef("Unmount All", self.unmountAll))

        return items

    def makeMenuItemsForMount(self, oid: Oid, widget: QWidget, single: bool = True):
        mc = self.mountedCommits[oid]

        items = [
            ActionDef(_("Open Mounted Folder"), mc.openFolder),
            ActionDef(_("Open in Terminal"), lambda mc=mc: mc.openTerminal(widget)),
            ActionDef(_("Copy Mount Point Path"), lambda mc=mc: self.copyPath(mc)),
            ActionDef(_("Unmount"), lambda mc=mc: self.unmount(mc)),
        ]

        if single:  # Single item: insert separator before 'Unmount'
            items.insert(-1, ActionDef.SEPARATOR)

        return items

    def checkOnClose(self, parentWidget, closeCallback):
        numMounts = len(self.mountedCommits)
        if numMounts == 0:
            return True

        oldestMount = next(iter(self.mountedCommits.values()))

        title = _("Unmount before quitting")
        prompt = _n(
            "Do you want to unmount folder {c} before quitting?",
            "Do you want to unmount {n} commit folders before quitting? ({c}, etc.)",
            n=numMounts, c=tquo(escamp(oldestMount.friendlyName())))

        buttons = QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel
        messageBox = asyncMessageBox(parentWidget, "question", title, prompt, buttons)

        quitButton = messageBox.button(QMessageBox.StandardButton.Ok)
        quitButton.setText(_("&Unmount && Quit"))
        quitButton.clicked.connect(lambda: self.unmountAll(closeCallback))

        messageBox.show()
        return False



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import logging
import re
from pathlib import Path
from collections.abc import Iterator

from gitfourchette.gitdriver.gitdelta import GitDelta, GitStatus
from gitfourchette.gitdriver.gitdeltafile import GitDeltaFile, GitDeltaSource, FileMode, HexHash0000, HexHashFFFF
from gitfourchette.gitdriver.gitconflict import GitConflict, GitConflictSides

_logger = logging.getLogger(__name__)


# 1 <XY> <sub> <mH> <mI> <mW> <hH> <hI> <path>
# 2 <XY> <sub> <mH> <mI> <mW> <hH> <hI> <R|C><score> <path><sep><origPath>
# u <XY> <sub> <m1> <m2> <m3> <mW> <h1> <h2> <h3> <path>
_gitStatusPatterns = {
    "1": re.compile(r"1 (.)(.) (....) (\d+) (\d+) (\d+) ([\da-f]+) ([\da-f]+) ([^\x00]*)\x00"),
    "2": re.compile(r"2 (.)(.) (....) (\d+) (\d+) (\d+) ([\da-f]+) ([\da-f]+) [RC](\d+) ([^\x00]*)\x00([^\x00]*)\x00"),
    "u": re.compile(r"u (..) (....) (\d+) (\d+) (\d+) (\d+) ([\da-f]+) ([\da-f]+) ([\da-f]+) ([^\x00]*)\x00"),
    "?": re.compile(r"\? ([^\x00]*)\x00"),
    "!": re.compile(r"! ([^\x00]*)\x00"),
}

_gitShowPattern = re.compile(r":(\d+) (\d+) ([\da-f]+) ([\da-f]+) (.)(\d*)\x00([^\x00]*)\x00")

# The order of this table is SIGNIFICANT!
_gitSimplifiedModes = [
    FileMode.LINK,              # 0o120000
    FileMode.TREE,              # 0o040000
    FileMode.BLOB_EXECUTABLE,   # 0o100755
    FileMode.BLOB,              # 0o100644
]

_aheadBehindPattern = re.compile(r"(.+) \[(?:behind (\d+)|ahead (\d+), behind (\d+)|ahead (\d+))]")


def distillMode(realMode: int) -> FileMode:
    """
    Git uses simplified file modes that may not accurately reflect a file's
    actual mode in the filesystem (e.g., a symlink's st_mode might be
    0o120777, which to git is just 0o120000). Use this function to simplify
    a real file mode to a legal FileMode value for git.
    """
    for gitMode in _gitSimplifiedModes:
        stripped = gitMode & ~0o000077  # Keep 'user' bits, ignore 'group'/'all' bits
        if stripped == (realMode & stripped):
            return gitMode

    _logger.warning(f"cannot map to git FileMode: 0o{realMode:o}")
    return FileMode.UNREADABLE


def parseMode(octal: str) -> FileMode:
    return FileMode(int(octal, 8))


def iterateLines(text: str) -> Iterator[tuple[int, int]]:
    pos = 0
    limit = len(text)

    while pos < limit:
        nextPos = text.find('\n', pos)
        if nextPos < 0:
            nextPos = limit
        else:
            nextPos += 1
        yield pos, nextPos
        pos = nextPos


def parseGitStatus(stdout: str, workdir: str) -> Iterator[tuple[GitDelta | None, GitDelta | None]]:
    pos = 0
    limit = len(stdout)

    while pos < limit:
        ident = stdout[pos]

        try:
            pattern = _gitStatusPatterns[ident]
        except KeyError as ex:
            raise ValueError(f"unknown ident {ident}") from ex

        match = pattern.match(stdout, pos)
        if match is None:
            raise ValueError(f"malformed record {ident}")

        pos = match.end()
        assert stdout[pos-1] == '\0'

        staged, unstaged = _parseStatusLine(ident, *match.groups())

        # Fill in file mode for untracked/ignored files.
        if (unstaged
                and unstaged.status in [GitStatus.Untracked, GitStatus.Ignored]
                and unstaged.new.mode == FileMode.UNREADABLE):
            try:
                stat = Path(workdir, unstaged.new.path).lstat()
                unstaged.new.mode = distillMode(stat.st_mode)
            except OSError:
                pass

        yield staged, unstaged


def _parseStatusLine(ident: str, *tokens: str) -> tuple[GitDelta | None, GitDelta | None]:
    if ident == "1":
        # Ordinary changed entries
        path = tokens[-1]
        tweakedTokens = tokens[:-1] + ("0", path, path)
        return _parseStatus2(*tweakedTokens)
    elif ident == "2":
        # Renamed or copied entries
        return _parseStatus2(*tokens)
    elif ident == "u":
        # Unmerged entries (conflict)
        return _parseStatusConflict(*tokens)
    elif ident == "?":
        # Untracked
        return _parseStatusUntracked(GitStatus.Untracked, *tokens)
    elif ident == "!":
        # Ignored
        return _parseStatusUntracked(GitStatus.Ignored, *tokens)
    else:
        raise ValueError(f"unknown ident: {ident}")


def _parseStatus2(x, y, sub, mh, mi, mw, hh, hi, score, newPath, origPath):
    fileHead = GitDeltaFile(
        path=origPath,
        id=hh,
        mode=parseMode(mh),
        source=GitDeltaSource.Commit)

    fileIndex = GitDeltaFile(
        path=newPath,
        id=hi,
        mode=parseMode(mi),
        source=GitDeltaSource.Index)

    fileWorktree = GitDeltaFile(
        path=newPath,
        id=HexHash0000 if y == "D" else HexHashFFFF,
        mode=parseMode(mw),
        source=GitDeltaSource.Dirty)

    xDelta, yDelta = None, None

    if x != ".":  # STAGED
        xStatus = GitStatus(x)
        xDelta = GitDelta(status=xStatus, old=fileHead, new=fileIndex, similarity=int(score))

    if y != ".":  # UNSTAGED
        yStatus = GitStatus(y)
        yDelta = GitDelta(status=yStatus, old=fileIndex, new=fileWorktree, submoduleStatus=sub)

    return xDelta, yDelta


def _parseStatusConflict(xy, sub, m1, m2, m3, mw, h1, h2, h3, path):
    indexFile = GitDeltaFile(path=path, source=GitDeltaSource.Index)
    worktreeFile = GitDeltaFile(path=path, id=HexHashFFFF, mode=parseMode(mw), source=GitDeltaSource.Dirty)

    sides = GitConflictSides(xy)
    stage1 = GitDeltaFile(path=path, id=h1, mode=parseMode(m1))
    stage2 = GitDeltaFile(path=path, id=h2, mode=parseMode(m2))
    stage3 = GitDeltaFile(path=path, id=h3, mode=parseMode(m3))
    conflict = GitConflict(sides, stage1, stage2, stage3)

    yDelta = GitDelta(status=GitStatus.Unmerged, old=indexFile, new=worktreeFile,
                      conflict=conflict, submoduleStatus=sub)
    return None, yDelta


def _parseStatusUntracked(status: GitStatus, path: str):
    if path.endswith("/"):
        path = path.removesuffix("/")
        mode = FileMode.TREE
    else:
        mode = FileMode.UNREADABLE  # a more precise mode will be filled in from the file's stats

    # "Old" state = empty file (not indexed yet)
    indexFile = GitDeltaFile(path=path, id=HexHash0000, mode=FileMode.UNREADABLE, source=GitDeltaSource.Index)

    worktreeFile = GitDeltaFile(path=path, id=HexHashFFFF, mode=mode, source=GitDeltaSource.Dirty)

    yDelta = GitDelta(status=status, old=indexFile, new=worktreeFile)
    return None, yDelta


def parseGitDiffRawZ(stdout: str):
    pos = 0
    limit = len(stdout)

    while pos < limit:
        match = _gitShowPattern.match(stdout, pos)
        pos = match.end()

        ms, md, hs, hd, status, score, path1 = match.groups()

        # WARNING! In case of a rename, "git show" outputs the old/new
        # paths in the reverse order from "git status --porcelain=v2"!
        # git show: ... old, new
        # git status: ... new, old
        if status in "RC":
            pos2 = stdout.find("\0", pos)
            path2 = stdout[pos:pos2]
            pos = pos2 + 1
        else:
            path2 = path1

        yield _parseShowLine(ms, md, hs, hd, status, score, path1, path2)


def _parseShowLine(ms, md, hs, hd, statusChar, score, path1, path2) -> GitDelta:
    status = GitStatus(statusChar)

    fileSrc = GitDeltaFile(
        path=path1,
        id=hs,
        mode=parseMode(ms),
        source=GitDeltaSource.Commit)

    fileDst = GitDeltaFile(
        path=path2,
        id=hd,
        mode=parseMode(md),
        source=GitDeltaSource.Commit)

    return GitDelta(status, fileSrc, fileDst, similarity=int(score) if score else 0)


def parseGitBlame(stdout: str) -> Iterator[tuple[str, int, str]]:
    # Transient data for current line
    commitId = ""
    originalLineNumber = -1

    for pos, endPos in iterateLines(stdout):
        if not commitId:  # Looking for header
            tokens = stdout[pos : endPos].split(" ", 2)
            commitId = tokens[0]
            originalLineNumber = int(tokens[1])
        elif stdout[pos] == '\t':
            text = stdout[pos+1 : endPos]
            yield commitId, originalLineNumber, text
            commitId = ""  # Look for next line
        else:
            # Ignore author, author-mail, etc.
            pass


def parseAheadBehind(stdout: str) -> Iterator[tuple[str, tuple[int, int]]]:
    for pos, endPos in iterateLines(stdout):
        match = _aheadBehindPattern.match(stdout, pos, endPos)
        if not match:
            continue
        ref, behindOnly, dualAhead, dualBehind, aheadOnly = match.groups()
        if behindOnly:
            yield ref, (0, int(behindOnly))
        elif aheadOnly:
            yield ref, (int(aheadOnly), 0)
        else:
            yield ref, (int(dualAhead), int(dualBehind))



# Form implementation generated from reading ui file 'newbranchdialog.ui'
#
# Created by: PyQt6 UI code generator 6.10.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_NewBranchDialog(object):
    def setupUi(self, NewBranchDialog):
        NewBranchDialog.setObjectName("NewBranchDialog")
        NewBranchDialog.setWindowModality(Qt.WindowModality.NonModal)
        NewBranchDialog.setEnabled(True)
        NewBranchDialog.resize(543, 202)
        NewBranchDialog.setSizeGripEnabled(False)
        NewBranchDialog.setModal(True)
        self.formLayout = QFormLayout(NewBranchDialog)
        self.formLayout.setFieldGrowthPolicy(QFormLayout.FieldGrowthPolicy.AllNonFixedFieldsGrow)
        self.formLayout.setObjectName("formLayout")
        self.nameLabel = QLabel(parent=NewBranchDialog)
        self.nameLabel.setObjectName("nameLabel")
        self.formLayout.setWidget(0, QFormLayout.ItemRole.LabelRole, self.nameLabel)
        self.optionsLabel = QLabel(parent=NewBranchDialog)
        self.optionsLabel.setObjectName("optionsLabel")
        self.formLayout.setWidget(2, QFormLayout.ItemRole.LabelRole, self.optionsLabel)
        self.switchToBranchCheckBox = QCheckBox(parent=NewBranchDialog)
        self.switchToBranchCheckBox.setChecked(True)
        self.switchToBranchCheckBox.setObjectName("switchToBranchCheckBox")
        self.formLayout.setWidget(2, QFormLayout.ItemRole.FieldRole, self.switchToBranchCheckBox)
        self.upstreamLayout = QHBoxLayout()
        self.upstreamLayout.setObjectName("upstreamLayout")
        self.upstreamCheckBox = QCheckBox(parent=NewBranchDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.upstreamCheckBox.sizePolicy().hasHeightForWidth())
        self.upstreamCheckBox.setSizePolicy(sizePolicy)
        self.upstreamCheckBox.setObjectName("upstreamCheckBox")
        self.upstreamLayout.addWidget(self.upstreamCheckBox)
        self.upstreamComboBox = QComboBox(parent=NewBranchDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.upstreamComboBox.sizePolicy().hasHeightForWidth())
        self.upstreamComboBox.setSizePolicy(sizePolicy)
        self.upstreamComboBox.setInsertPolicy(QComboBox.InsertPolicy.NoInsert)
        self.upstreamComboBox.setObjectName("upstreamComboBox")
        self.upstreamLayout.addWidget(self.upstreamComboBox)
        self.formLayout.setLayout(4, QFormLayout.ItemRole.FieldRole, self.upstreamLayout)
        self.buttonBox = QDialogButtonBox(parent=NewBranchDialog)
        self.buttonBox.setOrientation(Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Cancel|QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setCenterButtons(False)
        self.buttonBox.setObjectName("buttonBox")
        self.formLayout.setWidget(7, QFormLayout.ItemRole.FieldRole, self.buttonBox)
        self.nameEdit = QLineEdit(parent=NewBranchDialog)
        self.nameEdit.setObjectName("nameEdit")
        self.formLayout.setWidget(0, QFormLayout.ItemRole.FieldRole, self.nameEdit)
        self.recurseSubmodulesCheckBox = QCheckBox(parent=NewBranchDialog)
        self.recurseSubmodulesCheckBox.setChecked(True)
        self.recurseSubmodulesCheckBox.setObjectName("recurseSubmodulesCheckBox")
        self.formLayout.setWidget(3, QFormLayout.ItemRole.FieldRole, self.recurseSubmodulesCheckBox)
        self.nameLabel.setBuddy(self.nameEdit)

        self.retranslateUi(NewBranchDialog)
        self.buttonBox.rejected.connect(NewBranchDialog.reject)
        self.buttonBox.accepted.connect(NewBranchDialog.accept)
        self.upstreamCheckBox.toggled['bool'].connect(self.upstreamComboBox.setEnabled)
        self.switchToBranchCheckBox.toggled['bool'].connect(self.recurseSubmodulesCheckBox.setEnabled)
        NewBranchDialog.setTabOrder(self.nameEdit, self.switchToBranchCheckBox)
        NewBranchDialog.setTabOrder(self.switchToBranchCheckBox, self.recurseSubmodulesCheckBox)
        NewBranchDialog.setTabOrder(self.recurseSubmodulesCheckBox, self.upstreamCheckBox)
        NewBranchDialog.setTabOrder(self.upstreamCheckBox, self.upstreamComboBox)

    def retranslateUi(self, NewBranchDialog):
        NewBranchDialog.setWindowTitle(_p("NewBranchDialog", "New branch"))
        self.nameLabel.setText(_p("NewBranchDialog", "&Name:"))
        self.optionsLabel.setText(_p("NewBranchDialog", "Options:"))
        self.switchToBranchCheckBox.setText(_p("NewBranchDialog", "&Switch to branch after creating"))
        self.upstreamCheckBox.setText(_p("NewBranchDialog", "&Track upstream branch:"))
        self.recurseSubmodulesCheckBox.setText(_p("NewBranchDialog", "…then update submodules recursively"))



# Form implementation generated from reading ui file 'commitdialog.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_CommitDialog(object):
    def setupUi(self, CommitDialog):
        CommitDialog.setObjectName("CommitDialog")
        CommitDialog.resize(512, 208)
        CommitDialog.setModal(True)
        self.verticalLayout = QVBoxLayout(CommitDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.infoBox = QGroupBox(parent=CommitDialog)
        self.infoBox.setObjectName("infoBox")
        self.horizontalLayout_2 = QHBoxLayout(self.infoBox)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.infoIcon = QLabel(parent=self.infoBox)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.infoIcon.sizePolicy().hasHeightForWidth())
        self.infoIcon.setSizePolicy(sizePolicy)
        self.infoIcon.setMaximumSize(QSize(24, 16777215))
        self.infoIcon.setText("ICON")
        self.infoIcon.setAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignTop)
        self.infoIcon.setObjectName("infoIcon")
        self.horizontalLayout_2.addWidget(self.infoIcon)
        self.infoText = QLabel(parent=self.infoBox)
        self.infoText.setText("Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.")
        self.infoText.setWordWrap(True)
        self.infoText.setObjectName("infoText")
        self.horizontalLayout_2.addWidget(self.infoText)
        self.verticalLayout.addWidget(self.infoBox)
        self.summaryRowLayout = QHBoxLayout()
        self.summaryRowLayout.setObjectName("summaryRowLayout")
        self.summaryEditor = QLineEdit(parent=CommitDialog)
        self.summaryEditor.setObjectName("summaryEditor")
        self.summaryRowLayout.addWidget(self.summaryEditor)
        self.counterLabel = QLabel(parent=CommitDialog)
        self.counterLabel.setEnabled(False)
        self.counterLabel.setText("000")
        self.counterLabel.setObjectName("counterLabel")
        self.summaryRowLayout.addWidget(self.counterLabel)
        self.verticalLayout.addLayout(self.summaryRowLayout)
        self.descriptionEditor = QPlainTextEdit(parent=CommitDialog)
        self.descriptionEditor.setTabChangesFocus(True)
        self.descriptionEditor.setObjectName("descriptionEditor")
        self.verticalLayout.addWidget(self.descriptionEditor)
        self.signatureBox = QGroupBox(parent=CommitDialog)
        self.signatureBox.setEnabled(True)
        self.signatureBox.setFlat(False)
        self.signatureBox.setCheckable(False)
        self.signatureBox.setObjectName("signatureBox")
        self.verticalLayout_2 = QVBoxLayout(self.signatureBox)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.signature = SignatureForm(parent=self.signatureBox)
        self.signature.setObjectName("signature")
        self.verticalLayout_2.addWidget(self.signature)
        self.verticalLayout.addWidget(self.signatureBox)
        self.buttonLayout = QHBoxLayout()
        self.buttonLayout.setObjectName("buttonLayout")
        self.revealSignature = QCheckBox(parent=CommitDialog)
        self.revealSignature.setChecked(False)
        self.revealSignature.setObjectName("revealSignature")
        self.buttonLayout.addWidget(self.revealSignature)
        self.signatureButton = QHintButton(parent=CommitDialog)
        self.signatureButton.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.signatureButton.setAutoRaise(True)
        self.signatureButton.setObjectName("signatureButton")
        self.buttonLayout.addWidget(self.signatureButton)
        spacerItem = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.buttonLayout.addItem(spacerItem)
        self.hookButton = CommitHookButton(parent=CommitDialog)
        self.hookButton.setText("Hooks")
        self.hookButton.setObjectName("hookButton")
        self.buttonLayout.addWidget(self.hookButton)
        self.signoffButton = SignoffButton(parent=CommitDialog)
        self.signoffButton.setText("SOB")
        self.signoffButton.setObjectName("signoffButton")
        self.buttonLayout.addWidget(self.signoffButton)
        self.gpg = GpgButton(parent=CommitDialog)
        self.gpg.setText("&GPG")
        self.gpg.setObjectName("gpg")
        self.buttonLayout.addWidget(self.gpg)
        self.buttonBox = QDialogButtonBox(parent=CommitDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.buttonBox.sizePolicy().hasHeightForWidth())
        self.buttonBox.setSizePolicy(sizePolicy)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Cancel|QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.buttonLayout.addWidget(self.buttonBox)
        self.verticalLayout.addLayout(self.buttonLayout)

        self.retranslateUi(CommitDialog)
        self.buttonBox.accepted.connect(CommitDialog.accept)
        self.buttonBox.rejected.connect(CommitDialog.reject)
        self.revealSignature.toggled['bool'].connect(self.signatureBox.setVisible)
        self.revealSignature.toggled['bool'].connect(self.signatureBox.setEnabled)
        CommitDialog.setTabOrder(self.summaryEditor, self.descriptionEditor)

    def retranslateUi(self, CommitDialog):
        CommitDialog.setWindowTitle(_p("CommitDialog", "Commit"))
        self.summaryEditor.setPlaceholderText(_p("CommitDialog", "Enter commit summary"))
        self.counterLabel.setToolTip(_p("CommitDialog", "<p>Length of the summary. Convention is to stay under 50 characters and avoid going over 72. If that’s not enough, use the long-form description.</p>"))
        self.descriptionEditor.setPlaceholderText(_p("CommitDialog", "Long-form description (optional)"))
        self.revealSignature.setText(_p("CommitDialog", "&Edit Author"))
        self.signatureButton.setText(_p("CommitDialog", "Preview"))
from gitfourchette.forms.commithookbutton import CommitHookButton
from gitfourchette.forms.gpgbutton import GpgButton
from gitfourchette.forms.signatureform import SignatureForm
from gitfourchette.forms.signoffbutton import SignoffButton
from gitfourchette.toolbox.qhintbutton import QHintButton



# Form implementation generated from reading ui file 'identitydialog.ui'
#
# Created by: PyQt6 UI code generator 6.10.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_IdentityDialog(object):
    def setupUi(self, IdentityDialog):
        IdentityDialog.setObjectName("IdentityDialog")
        IdentityDialog.resize(465, 269)
        self.formLayout = QFormLayout(IdentityDialog)
        self.formLayout.setFieldGrowthPolicy(QFormLayout.FieldGrowthPolicy.ExpandingFieldsGrow)
        self.formLayout.setObjectName("formLayout")
        self.nameLabel = QLabel(parent=IdentityDialog)
        self.nameLabel.setObjectName("nameLabel")
        self.formLayout.setWidget(0, QFormLayout.ItemRole.LabelRole, self.nameLabel)
        self.nameEdit = QLineEdit(parent=IdentityDialog)
        self.nameEdit.setObjectName("nameEdit")
        self.formLayout.setWidget(0, QFormLayout.ItemRole.FieldRole, self.nameEdit)
        self.emailLabel = QLabel(parent=IdentityDialog)
        self.emailLabel.setObjectName("emailLabel")
        self.formLayout.setWidget(1, QFormLayout.ItemRole.LabelRole, self.emailLabel)
        self.emailEdit = QLineEdit(parent=IdentityDialog)
        self.emailEdit.setObjectName("emailEdit")
        self.formLayout.setWidget(1, QFormLayout.ItemRole.FieldRole, self.emailEdit)
        self.buttonBox = QDialogButtonBox(parent=IdentityDialog)
        self.buttonBox.setOrientation(Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Cancel|QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.formLayout.setWidget(7, QFormLayout.ItemRole.SpanningRole, self.buttonBox)
        self.warningLabel = QLabel(parent=IdentityDialog)
        self.warningLabel.setEnabled(False)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.warningLabel.sizePolicy().hasHeightForWidth())
        self.warningLabel.setSizePolicy(sizePolicy)
        self.warningLabel.setWordWrap(True)
        self.warningLabel.setObjectName("warningLabel")
        self.formLayout.setWidget(4, QFormLayout.ItemRole.SpanningRole, self.warningLabel)
        self.configPathLabel = QLabel(parent=IdentityDialog)
        self.configPathLabel.setEnabled(False)
        self.configPathLabel.setObjectName("configPathLabel")
        self.formLayout.setWidget(3, QFormLayout.ItemRole.SpanningRole, self.configPathLabel)
        spacerItem = QSpacerItem(20, 1, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.formLayout.setItem(2, QFormLayout.ItemRole.LabelRole, spacerItem)
        self.nameLabel.setBuddy(self.nameEdit)
        self.emailLabel.setBuddy(self.emailEdit)

        self.retranslateUi(IdentityDialog)
        self.buttonBox.accepted.connect(IdentityDialog.accept)
        self.buttonBox.rejected.connect(IdentityDialog.reject)
        IdentityDialog.setTabOrder(self.nameEdit, self.emailEdit)

    def retranslateUi(self, IdentityDialog):
        IdentityDialog.setWindowTitle(_p("IdentityDialog", "Set up your Git identity"))
        self.nameLabel.setText(_p("IdentityDialog", "&Name:"))
        self.emailLabel.setText(_p("IdentityDialog", "&Email:"))
        self.warningLabel.setText(_p("IdentityDialog", "Note: This will not apply to the current repository, because its configuration file defines a custom identity."))
        self.configPathLabel.setText(_p("IdentityDialog", "This information will be saved in {0}."))



# NOTE: Some parts adapted from cpython

import glob as _glob
import sys as _sys
import typing as _typing
from pathlib import Path as _Path


# Python 3.12 compatibility
# (`follow_symlinks` argument in Path.is_file() is new in Python 3.13)
if _sys.version_info < (3, 13):
    def _pathIsFile(self: _Path, follow_symlinks=True) -> bool:
        if follow_symlinks:
            return _pathIsFileVanilla(self)
        return not self.is_symlink() and _pathIsFileVanilla(self)

    _pathIsFileVanilla = _Path.is_file
    _Path.is_file = _pathIsFile  # type: ignore[method-assign]


# Python 3.12 compatibility (glob.translate is new Python 3.13)
# Adapted from cpython/Lib/glob.py
if not hasattr(_glob, "translate"):
    def _globTranslate(pat: str, recursive: bool, include_hidden: bool):
        if not recursive:
            raise NotImplementedError()
        if not include_hidden:
            raise NotImplementedError()

        import re
        seps = "/"
        escaped_seps = ''.join(map(re.escape, seps))
        any_sep = escaped_seps
        not_sep = f'[^{escaped_seps}]'
        one_last_segment = f'{not_sep}+'
        one_segment = f'{one_last_segment}{any_sep}'
        any_segments = f'(?:.+{any_sep})?'
        any_last_segments = '.*'

        results = []
        parts = re.split(any_sep, pat)
        last_part_idx = len(parts) - 1
        for idx, part in enumerate(parts):
            if part == '*':
                results.append(one_segment if idx < last_part_idx else one_last_segment)
            elif part == '**':
                if idx < last_part_idx:
                    if parts[idx + 1] != '**':
                        results.append(any_segments)
                else:
                    results.append(any_last_segments)
            else:
                if part:
                    results.extend(_fnmatch_translate(part, f'{not_sep}*', not_sep))
                if idx < last_part_idx:
                    results.append(any_sep)
        res = ''.join(results)
        return fr'(?s:{res})\Z'

    @_typing.no_type_check
    def _fnmatch_translate(pat, star, question_mark):
        import re
        import functools
        _re_setops_sub = re.compile(r'([&~|])').sub
        _re_escape = functools.lru_cache(maxsize=512)(re.escape)

        res = []
        add = res.append

        i, n = 0, len(pat)
        while i < n:
            c = pat[i]
            i = i + 1
            if c == '*':
                # store the position of the wildcard
                add(star)
                # compress consecutive `*` into one
                while i < n and pat[i] == '*':
                    i += 1
            elif c == '?':
                add(question_mark)
            elif c == '[':
                j = i
                if j < n and pat[j] == '!':
                    j = j + 1
                if j < n and pat[j] == ']':
                    j = j + 1
                while j < n and pat[j] != ']':
                    j = j + 1
                if j >= n:
                    add('\\[')
                else:
                    stuff = pat[i:j]
                    if '-' not in stuff:
                        stuff = stuff.replace('\\', r'\\')
                    else:
                        chunks = []
                        k = i + 2 if pat[i] == '!' else i + 1
                        while True:
                            k = pat.find('-', k, j)
                            if k < 0:
                                break
                            chunks.append(pat[i:k])
                            i = k + 1
                            k = k + 3
                        chunk = pat[i:j]
                        if chunk:
                            chunks.append(chunk)
                        else:
                            chunks[-1] += '-'
                        # Remove empty ranges -- invalid in RE.
                        for k in range(len(chunks) - 1, 0, -1):
                            if chunks[k - 1][-1] > chunks[k][0]:
                                chunks[k - 1] = chunks[k - 1][:-1] + chunks[k][1:]
                                del chunks[k]
                        # Escape backslashes and hyphens for set difference (--).
                        # Hyphens that create ranges shouldn't be escaped.
                        stuff = '-'.join(s.replace('\\', r'\\').replace('-', r'\-')
                                         for s in chunks)
                    i = j + 1
                    if not stuff:
                        # Empty range: never match.
                        add('(?!)')
                    elif stuff == '!':
                        # Negated empty range: match any character.
                        add('.')
                    else:
                        # Escape set operations (&&, ~~ and ||).
                        stuff = _re_setops_sub(r'\\\1', stuff)
                        if stuff[0] == '!':
                            stuff = '^' + stuff[1:]
                        elif stuff[0] in ('^', '['):
                            stuff = '\\' + stuff
                        add(f'[{stuff}]')
            else:
                add(_re_escape(c))
        assert i == n
        return res

    _glob.translate = _globTranslate  # type: ignore[assignment]



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from typing import ClassVar

from gitfourchette.qt import *
from gitfourchette.toolbox import MultiShortcut, makeMultiShortcut


class GlobalShortcuts:
    NO_SHORTCUT: ClassVar[MultiShortcut] = []

    find: MultiShortcut = NO_SHORTCUT
    findNext: MultiShortcut = NO_SHORTCUT
    findPrevious: MultiShortcut = NO_SHORTCUT
    refresh: MultiShortcut = NO_SHORTCUT
    openRepoFolder: MultiShortcut = NO_SHORTCUT
    openTerminal: MultiShortcut = NO_SHORTCUT

    stageHotkeys: ClassVar = [Qt.Key.Key_Return, Qt.Key.Key_Enter]  # Return: main keys; Enter: on keypad
    discardHotkeys: ClassVar = [Qt.Key.Key_Delete, Qt.Key.Key_Backspace]

    _initialized = False

    @classmethod
    def initialize(cls):
        assert QApplication.instance(), "QApplication must have been created before instantiating QKeySequence"
        assert not cls._initialized, "GlobalShortcuts already initialized"

        # Non-KDE environments (incl. GNOME, Cinnamon, Windows) typically bind
        # FindNext to Ctrl+G, but this shortcut should be reserved for
        # JumpToUncommittedChanges (except on macOS where we use Meta+G).
        overrideCtrlG = not MACOS

        cls.find = makeMultiShortcut(QKeySequence.StandardKey.Find, "/")
        cls.findNext = makeMultiShortcut("F3" if overrideCtrlG else QKeySequence.StandardKey.FindNext)
        cls.findPrevious = makeMultiShortcut("Shift+F3" if overrideCtrlG else QKeySequence.StandardKey.FindPrevious)
        cls.refresh = makeMultiShortcut(QKeySequence.StandardKey.Refresh, "Ctrl+R", "F5")
        cls.openRepoFolder = makeMultiShortcut("Ctrl+Shift+O")
        cls.openTerminal = makeMultiShortcut("Ctrl+Alt+O")

        cls._initialized = True



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import os
import shlex
from typing import ClassVar

from gitfourchette.exttools.toolcommands import ToolCommands
from gitfourchette.qt import *


class ToolPresets:
    DefaultDiffCommand = ""
    DefaultMergeCommand = ""
    DefaultTerminalCommand = ""
    FlatpakNamePrefix = "Flatpak: "

    Editors: ClassVar = {
        "System default": "",
        "BBEdit"        : "bbedit",
        "GVim"          : "gvim",
        "Kate"          : "kate",
        "KWrite"        : "kwrite",
        "MacVim"        : "mvim",
        "VS Code"       : "code",
    }

    DiffTools: ClassVar = {
        "Beyond Compare": "bcompare $L $R",
        "CLion"         : "clion diff $L $R",
        "DiffMerge"     : "diffmerge $L $R",
        "FileMerge"     : "assets:mac/opendiff.sh $L $R",
        "GVim"          : "gvim -f -d $L $R",
        "IntelliJ IDEA" : "idea diff $L $R",
        "KDiff3"        : "kdiff3 $L $R",
        "MacVim"        : "mvim -f -d $L $R",
        "Meld"          : "meld $L $R",
        "P4Merge"       : "p4merge $L $R",
        "PyCharm"       : "pycharm diff $L $R",
        "VS Code"       : "code --new-window --wait --diff $L $R",
        "WinMerge"      : "winmergeu /u /wl /wr $L $R",
    }

    # $B: ANCESTOR/BASE/CENTER
    # $L: OURS/LOCAL/LEFT
    # $R: THEIRS/REMOTE/RIGHT
    # $M: MERGED/OUTPUT
    MergeTools: ClassVar = {
        "Beyond Compare": "bcompare $L $R $B $M",
        "CLion"         : "clion merge $L $R $B $M",
        "DiffMerge"     : "diffmerge --merge --result=$M $L $B $R",
        "FileMerge"     : "assets:mac/opendiff.sh -ancestor $B $L $R -merge $M",
        "GVim"          : "gvim -f -d -c 'wincmd J' $M $L $B $R",
        "IntelliJ IDEA" : "idea merge $L $R $B $M",
        "KDiff3"        : "kdiff3 --merge $B $L $R --output $M",
        "MacVim"        : "mvim -f -d -c 'wincmd J' $M $L $B $R",
        "Meld"          : "meld --auto-merge $L $B $R --output=$M",
        "P4Merge"       : "p4merge $B $L $R $M",
        "PyCharm"       : "pycharm merge $L $R $B $M",
        "VS Code"       : "code --new-window --wait --merge $L $R $B $M",
        "WinMerge"      : "winmergeu /u /wl /wm /wr /am $B $L $R /o $M",
    }

    _macTerminals: ClassVar = {
        "macOS Terminal": "assets:mac/terminal.scpt $COMMAND",
        "kitty"         : "kitty --single-instance $COMMAND",  # single instance looks better in dock
        "WezTerm"       : "wezterm start $COMMAND",  # 'start' instead of '-e' to reuse app instance
        "Ghostty"       : "open -na ghostty.app --args -e $COMMAND",
    }

    _windowsTerminals: ClassVar = {
        # By default, git-bash starts bash with -i --login, which appends the
        # contents of our kicker script to the user's .bash_history.
        # To bypass this, invoke mintty manually.
        "Git Bash"      : "git-bash --command=usr/bin/mintty.exe --exec $COMMAND",
        # Can't use cmd or powershell on Windows for now, because we generate a bash script
        # that wraps the command to run.
        # "Command Prompt": "cmd /c start cmd /k $COMMAND",
        # "PowerShell"    : "cmd /c start powershell -NoExit -Command $COMMAND",
    }

    _freedesktopTerminals: ClassVar = {
        "Alacritty"     : "alacritty -e $COMMAND",
        "Contour"       : "contour $COMMAND",
        "foot"          : "foot $COMMAND",
        "Ghostty"       : "ghostty -e $COMMAND",
        "GNOME Terminal": "gnome-terminal -- $COMMAND",
        "kitty"         : "kitty $COMMAND",
        "Konsole"       : "konsole -e $COMMAND",
        "Ptyxis"        : "ptyxis -x $COMMAND",
        "st"            : "st -e $COMMAND",
        "Terminator"    : "terminator -x $COMMAND",
        "urxvt"         : "urxvt -e $COMMAND",
        "WezTerm"       : "wezterm -e $COMMAND",
        "xterm"         : "xterm -e $COMMAND",
    }

    # Filled in depending on platform
    Terminals: ClassVar[dict[str, str]] = {}

    FlatpakIDs: ClassVar = {
        "CLion"             : ("CLion",         "com.jetbrains.CLion"),
        "GVim"              : ("GVim",          "org.vim.Vim"),
        "IntelliJ IDEA CE"  : ("IntelliJ IDEA", "com.jetbrains.IntelliJ-IDEA-Community"),
        "PyCharm CE"        : ("PyCharm",       "com.jetbrains.PyCharm-Community"),
        "Kate"              : ("Kate",          "org.kde.kate"),
        "KDiff3"            : ("KDiff3",        "org.kde.kdiff3"),
        "KWrite"            : ("KWrite",        "org.kde.kwrite"),
        "Meld"              : ("Meld",          "org.gnome.meld"),
        "VS Code"           : ("VS Code",       "com.visualstudio.code"),
        "VS Code OSS"       : ("VS Code",       "com.visualstudio.code-oss"),
        # Terminals
        "Konsole"           : ("Konsole",       "org.kde.konsole"),
    }

    @classmethod
    def _filterToolPresets(cls):
        freedesktopTools = ["Kate", "KWrite"]
        macTools = ["FileMerge", "MacVim", "BBEdit"]
        winTools = ["WinMerge"]
        allPresetDicts = [cls.Editors, cls.DiffTools, cls.MergeTools, cls.Terminals]

        if MACOS:
            excludeTools = winTools + freedesktopTools
            cls.Terminals.update(cls._macTerminals)
            defaultDiffPreset = "FileMerge"
            defaultMergePreset = "FileMerge"
            defaultTerminalPreset = "macOS Terminal"
        elif WINDOWS:
            excludeTools = macTools + freedesktopTools
            cls.Terminals.update(cls._windowsTerminals)
            defaultDiffPreset = "WinMerge"
            defaultMergePreset = "WinMerge"
            defaultTerminalPreset = "Git Bash"
        else:
            excludeTools = macTools + winTools
            cls.Terminals.update(cls._freedesktopTerminals)

            terminalScores = dict.fromkeys(cls._freedesktopTerminals, 0)
            terminalScores |= {
                "Ptyxis"        : 2 if GNOME else -2,  # Fedora default
                "GNOME Terminal": 1 if GNOME else -1,
                "Konsole"       : 1 if KDE else 0,
                "xterm"         : -4 if WAYLAND else -3,
            }

            defaultDiffPreset = "Meld"
            defaultMergePreset = "Meld"
            defaultTerminalPreset = cls._findBestPreset(cls._freedesktopTerminals, terminalScores, "Konsole")

        for key in excludeTools:
            for presets in allPresetDicts:
                try:
                    del presets[key]
                except KeyError:
                    pass

        if FREEDESKTOP:
            for name, (alias, flatpakId) in cls.FlatpakIDs.items():
                k2 = cls.FlatpakNamePrefix + name
                assert any(
                    alias in presets for presets in allPresetDicts), f"missing non-flatpak preset for {alias}"
                for presets in allPresetDicts:
                    try:
                        originalCommand = presets[alias]
                    except KeyError:
                        continue
                    newCommand = ToolCommands.replaceProgramTokenInCommand(originalCommand, "flatpak", "run", flatpakId)
                    presets[k2] = newCommand

        cls.DefaultDiffCommand = cls._postProcessDefaultPreset(defaultDiffPreset, cls.DiffTools)
        cls.DefaultMergeCommand = cls._postProcessDefaultPreset(defaultMergePreset, cls.MergeTools)
        cls.DefaultTerminalCommand = cls._postProcessDefaultPreset(defaultTerminalPreset, cls.Terminals)

    @classmethod
    def _findBestPreset(cls, presets: dict[str, str], scores: dict[str, int], fallback: str) -> str:
        assert fallback in presets
        assert set(scores.keys()) == set(presets.keys()), "scores and presets must have the same keys!"

        sortedKeys = sorted(scores.keys(), key=lambda k: scores[k], reverse=True)

        for key in sortedKeys:
            program = presets[key].split()[0]
            if ToolCommands.which(program):
                return key

        return fallback

    @classmethod
    def _postProcessDefaultPreset(cls, key, presets):
        assert key in presets

        # If we're running as a Flatpak, use Flatpak as default tool as well
        if FLATPAK:
            flatpakKey = cls.FlatpakNamePrefix + key
            if flatpakKey in presets:
                key = flatpakKey

        return presets[key]

    @classmethod
    def getCommandName(cls, command: str, fallback = "", presets: dict[str, str] | None = None) -> str:
        if not command.strip():
            return fallback

        if presets is not None:
            presetName = next((k for k, v in presets.items() if v == command), "")
            if presetName:
                if presetName.startswith(cls.FlatpakNamePrefix):
                    presetName = presetName.removeprefix(cls.FlatpakNamePrefix)
                    presetName += " (Flatpak)"
                return presetName

        tokens = ToolCommands.splitCommandTokens(command)
        interestingToken = 0

        if FREEDESKTOP:
            interestingToken = ToolCommands.isFlatpakRunCommand(tokens)
            assert interestingToken >= 0

        try:
            name = tokens[interestingToken]
        except IndexError:
            return fallback

        name = os.path.basename(name)

        if MACOS:
            name = name.removesuffix(".app")

        return name

    @classmethod
    def flatpakBuiltInGit(cls) -> str:
        builtInGitPath = "/app/bin/git"
        if FLATPAK and os.path.exists(builtInGitPath):
            return ToolCommands.FlatpakSandboxedCommandPrefix + builtInGitPath
        return ""

    @classmethod
    def defaultGit(cls, hostOnly=False) -> str:
        return shlex.quote(
            (not hostOnly and cls.flatpakBuiltInGit())
            or ToolCommands.which("git")
            or "/usr/bin/git")


ToolPresets._filterToolPresets()



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations as _annotations

import configparser as _configparser
import dataclasses as _dataclasses
import datetime
import enum
import logging as _logging
import os as _os
import re as _re
import shutil as _shutil
import typing as _typing
import warnings
from contextlib import suppress as _suppress
from os.path import (
    abspath as _abspath,
    basename as _basename,
    dirname as _dirname,
    exists as _exists,
    isabs as _isabs,
    isdir as _isdir,
    isfile as _isfile,
    join as _joinpath,
    normpath as _normpath,
    relpath as _relpath,
)
from pathlib import Path as _Path

from pygit2 import settings as _pygit2settings

from pygit2 import (
    Blob,
    Branch,
    CheckoutCallbacks,
    Commit,
    Config as GitConfig,
    Diff,
    DiffDelta,
    DiffFile,
    DiffLine,
    DiffHunk,
    GitError,
    Keypair,
    IndexEntry,
    InvalidSpecError,
    Object as GitObject,
    Oid,
    Patch,
    Remote,
    RemoteCallbacks,
    Repository as _VanillaRepository,
    Signature,
    Stash,
    StashApplyCallbacks,
    Submodule,
    Tree,
    Walker,

    __version__ as PYGIT2_VERSION,
    LIBGIT2_VERSION,
)

from pygit2.enums import (
    ApplyLocation,
    BranchType,
    CheckoutNotify,
    CheckoutStrategy,
    ConfigLevel as GitConfigLevel,
    CredentialType,
    DeltaStatus,
    DiffFlag,
    DiffOption,
    FileStatus,
    FileMode,
    FetchPrune,
    MergeAnalysis,
    MergePreference,
    ObjectType,
    RepositoryOpenFlag,
    RepositoryState,
    ReferenceType,
    ResetMode,
    SortMode,
)

from pygit2.remotes import TransferProgress


_logger = _logging.getLogger(__name__)

NULL_OID = Oid(raw=b'')
"""
A hash consisting of all zeros.
Used in some places to denote the absence of an object.
"""

EMPTYTREE_OID = Oid(hex="4b825dc642cb6eb9a060e54bf8d69288fbee4904")
"""
SHA-1 hash of the empty tree (sha1(b"tree \0")).
"""

DOT_GITMODULES = ".gitmodules"

CORE_STASH_MESSAGE_PATTERN = _re.compile(r"^On (?:[^\s:]+|\(no branch\)): (.+)")
WINDOWS_RESERVED_FILENAMES_PATTERN = _re.compile(r"(.*/)?(AUX|COM[1-9]|CON|LPT[1-9]|NUL|PRN)($|\.|/)", _re.IGNORECASE)
DIFF_HEADER_PATTERN = _re.compile(r"^diff --git (\"?\w/[^\"]+\"?) (\"?\w/[^\"]+\"?)")

SUBPROJECT_COMMIT_MINUS_PATTERN = _re.compile(r"^-Subproject commit (.+)$", _re.MULTILINE)
SUBPROJECT_COMMIT_PLUS_PATTERN = _re.compile(r"^\+Subproject commit (.+)$", _re.MULTILINE)

FileStatus_INDEX_MASK = (
        FileStatus.INDEX_NEW
        | FileStatus.INDEX_MODIFIED
        | FileStatus.INDEX_DELETED
        | FileStatus.INDEX_RENAMED
        | FileStatus.INDEX_TYPECHANGE)

FileStatus_WT_MASK = (
        FileStatus.WT_NEW
        | FileStatus.WT_MODIFIED
        | FileStatus.WT_DELETED
        | FileStatus.WT_TYPECHANGE
        | FileStatus.WT_RENAMED
        | FileStatus.WT_UNREADABLE)


class RefPrefix:
    HEADS = "refs/heads/"
    REMOTES = "refs/remotes/"
    TAGS = "refs/tags/"

    @classmethod
    def split(cls, refname: str) -> tuple[str, str]:
        for prefix in cls.HEADS, cls.REMOTES, cls.TAGS:
            if refname.startswith(prefix):
                return prefix, refname[len(prefix):]
        return "", refname


class NameValidationError(ValueError):
    class Rule(enum.IntEnum):
        CANNOT_BE_EMPTY = 0
        ILLEGAL_NAME = 1
        ILLEGAL_PREFIX = 2
        ILLEGAL_SUFFIX = 3
        CONTAINS_ILLEGAL_CHAR = 4
        CONTAINS_ILLEGAL_SEQ = 5
        NOT_WINDOWS_FRIENDLY = 6
        NAME_TAKEN_BY_REF = 7
        NAME_TAKEN_BY_FOLDER = 8
        NOT_A_FOLDER = 9

    def __init__(self, rule: Rule):
        super().__init__(f"Name validation failed ({rule})")
        self.rule = rule


class DivergentBranchesError(Exception):
    def __init__(self, local_branch: Branch, remote_branch: Branch):
        super().__init__()
        self.local_branch = local_branch
        self.remote_branch = remote_branch

    def __repr__(self):
        return f"DivergentBranchesError(local: {self.local_branch.shorthand}, remote: {self.remote_branch.shorthand})"


class ConflictError(Exception):
    def __init__(self, conflicts: list[str], description="Conflicts"):
        super().__init__(description)
        self.description = description
        self.conflicts = conflicts

    def __repr__(self):
        return f"ConflictError({len(self.conflicts)}, {self.description})"


class MultiFileError(Exception):
    file_exceptions: dict[str, Exception | None]

    def __init__(self, message: str = ""):
        self.file_exceptions = {}
        self.message = message
        self.file_successes = 0

    def add_file_error(self, path: str, exc: Exception | None):
        self.file_exceptions[path] = exc

    def add_file_success(self):
        self.file_successes += 1

    def __bool__(self):
        return bool(self.file_exceptions)


class _UnchangedConfigFile(Exception):
    pass


class CheckoutBreakdown(CheckoutCallbacks):
    status: dict[str, CheckoutNotify]

    def __init__(self):
        super().__init__()
        self.status = {}

    def checkout_notify(self, why: CheckoutNotify, path: str, baseline=None, target=None, workdir=None):
        self.status[path] = why

    def get_conflicts(self):
        return [path for path in self.status
                if self.status[path] == CheckoutNotify.CONFLICT]

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not exc_type:
            return False
        if issubclass(exc_type, GitError):
            message = str(exc_val)
            if "prevents checkout" in message or "prevent checkout" in message:
                conflicts = self.get_conflicts()
                if conflicts:
                    raise ConflictError(conflicts, "workdir")
        return False  # let error propagate


def version_to_tuple(s: str) -> tuple[int, ...]:
    v = []
    for n in s.split("."):
        try:
            v.append(int(n))
        except ValueError:
            v.append(0)
    # Trim trailing zeros to ease comparison
    while v and v[-1] == 0:
        v.pop()
    return tuple(v)


def _version_at_least(
        package_name: str,
        required_version_string: str,
        current_version_string: str,
        raise_error=True,
        feature_name="This feature"
):
    required_version = version_to_tuple(required_version_string)
    current_version = version_to_tuple(current_version_string)

    if tuple(current_version) >= tuple(required_version):
        return True

    message = (f"{feature_name} requires {package_name} v{required_version_string} or later "
               f"(you have v{current_version_string}).")
    if raise_error:
        raise NotImplementedError(message)
    else:
        warnings.warn(message)
        return False


def pygit2_version_at_least(required_version: str, raise_error=True, feature_name="This feature"):
    return _version_at_least(
        package_name="pygit2",
        required_version_string=required_version,
        current_version_string=PYGIT2_VERSION,
        raise_error=raise_error,
        feature_name=feature_name)


def libgit2_version_at_least(required_version: str, raise_error=True, feature_name="This feature"):
    return _version_at_least(
        package_name="libgit2",
        required_version_string=required_version,
        current_version_string=LIBGIT2_VERSION,
        raise_error=raise_error,
        feature_name=feature_name)


def id7(obj: str | Oid | GitObject, n=7):
    """
    Shorten oid or Object oid
    """
    if isinstance(obj, str):
        sid = obj
    elif isinstance(obj, Oid):
        sid = str(obj)
    else:
        sid = str(obj.id)
    return sid[:n]


def commit_diff_pair(commit: Commit) -> tuple[Oid | None, Oid]:
    try:
        # Compare to first parent
        a = commit.parent_ids[0]
    except IndexError:
        # Root commit, compare to empty tree
        a = None

    return a, commit.id


def split_remote_branch_shorthand(remote_branch_name: str) -> tuple[str, str]:
    """
    Extract the remote name and branch name from a remote branch shorthand
    string such as "origin/master".

    The input string must not start with "refs/remotes/".

    Note: results can be flaky if the remote contains a slash in its name.
    """

    if remote_branch_name.startswith("refs/"):
        raise ValueError("remote branch shorthand name must not start with refs/")

    # TODO: extraction of branch name is flaky if remote name or branch name contains slashes
    try:
        remote_name, branch_name = remote_branch_name.split("/", 1)
        return remote_name, branch_name
    except ValueError:
        # `git svn clone` creates .git/refs/remotes/git-svn, which trips up pygit2
        return remote_branch_name, ""


def validate_refname(name: str, reserved_names: list[str], allow_dash_prefix: bool = False):
    """
    Checks the validity of a ref name according to `man git-check-ref-format`.
    Raises NameValidationError if the name is incorrect.
    """

    # Can't be empty
    if not name:
        raise NameValidationError(NameValidationError.Rule.CANNOT_BE_EMPTY)

    # Rule 9: can't be single character '@'
    if name == '@':
        raise NameValidationError(NameValidationError.Rule.ILLEGAL_NAME)

    # Rule 4: forbid space, tilde, caret, colon
    # Rule 5: forbid question mark, asterisk, open bracket
    # Rule 10: forbid backslash
    if any(c in " ~^:[?*\\" for c in name):
        raise NameValidationError(NameValidationError.Rule.CONTAINS_ILLEGAL_CHAR)

    # Rule 1: slash-separated components can't start with dot or end with .lock
    # Rule 3: forbid consecutive dots
    # Rule 6: forbid consecutive slashes
    # Rule 8: forbid '@{'
    if any(seq in name for seq in ["/.", ".lock/", "..", "//", "@{"]):
        raise NameValidationError(NameValidationError.Rule.CONTAINS_ILLEGAL_SEQ)

    # Rule 1: can't start with dot
    # Rule 6: can't start with slash
    if name.startswith((".", "/")):
        raise NameValidationError(NameValidationError.Rule.ILLEGAL_PREFIX)

    # Also: branches and tags can't start with dash
    if not allow_dash_prefix and name.startswith("-"):
        raise NameValidationError(NameValidationError.Rule.ILLEGAL_PREFIX)

    # Rule 1: can't end with .lock
    # Rule 6: can't end with slash
    # Rule 7: can't end with dot
    if name.endswith((".lock", "/", ".")):
        raise NameValidationError(NameValidationError.Rule.ILLEGAL_SUFFIX)

    # Prevent filenames that are reserved on Windows
    if WINDOWS_RESERVED_FILENAMES_PATTERN.match(name):
        raise NameValidationError(NameValidationError.Rule.NOT_WINDOWS_FRIENDLY)

    # Don't clash with existing refs
    if name.lower() in (n.lower() for n in reserved_names):
        raise NameValidationError(NameValidationError.Rule.NAME_TAKEN_BY_REF)

    # Don't clash with ref folders. If you attempt to rename a branch to the
    # name of an existing folder, libgit2 first deletes the branch, then errors
    # out with "cannot lock ref 'refs/heads/folder', there are refs beneath
    # that folder". So, let's avoid losing the branch.
    folder = name.lower() + "/"
    if any(n.lower().startswith(folder) for n in reserved_names):
        raise NameValidationError(NameValidationError.Rule.NAME_TAKEN_BY_FOLDER)

    # Prevent creating a folder with the same name as an existing ref
    if "/" in name and any(name.lower().startswith(n.lower() + "/") for n in reserved_names):
        raise NameValidationError(NameValidationError.Rule.NOT_A_FOLDER)


def validate_signature_item(s: str):
    """
    Checks the validity of the name or email in a signature according to `libgit2/signature.c`.
    Raises NameValidationError if the item is incorrect.
    """

    # Angle bracket characters are not allowed
    if "<" in s or ">" in s:
        raise NameValidationError(NameValidationError.Rule.CONTAINS_ILLEGAL_CHAR)

    # Trim crud from name
    def is_crud(c: str):
        return ord(c) <= 32 or c in ".,:;<>\"\\'"

    start = 0
    end = len(s)
    while end > 0 and is_crud(s[end-1]):
        end -= 1
    while start < end and is_crud(s[start]):
        start += 1
    trimmed = s[start:end]

    # Cannot be empty after trimming
    if not trimmed:
        raise NameValidationError(NameValidationError.Rule.CANNOT_BE_EMPTY)


def signatures_equalish(a: Signature, b: Signature) -> bool:
    """
    Sometimes two signature objects differ only by the encoding field.
    """

    if a == b:
        return True

    if not isinstance(a, Signature) or not isinstance(b, Signature):
        return False  # type: ignore[unreachable]

    if a._encoding == b._encoding:
        # If the encodings match and a != b, then the signatures are really different
        return False

    return (a.email == b.email
        and a.name == b.name
        and a.raw_email == b.raw_email
        and a.raw_name == b.raw_name
        and a.time == b.time
        and a.offset == b.offset
    )


def strip_stash_message(stash_message: str) -> str:
    m = CORE_STASH_MESSAGE_PATTERN.match(stash_message)
    if m:
        return m.group(1)
    # Remove linebreaks in WIP-type messages
    return stash_message.split('\n', maxsplit=1)[0]


def parse_submodule_patch(text: str) -> tuple[Oid, Oid, bool]:
    def parse_subproject_line(match: _re.Match | None):
        dirty = False
        if match is None:
            oid = NULL_OID
        else:
            capture = match.group(1)
            if capture.endswith("-dirty"):
                capture = capture.removesuffix("-dirty")
                dirty = True
            oid = Oid(hex=capture)
        return oid, dirty

    old_match = SUBPROJECT_COMMIT_MINUS_PATTERN.search(text)
    new_match = SUBPROJECT_COMMIT_PLUS_PATTERN.search(text)
    old_id, _dummy = parse_subproject_line(old_match)
    new_id, new_dirty = parse_subproject_line(new_match)
    return old_id, new_id, new_dirty


class GitConfigHelper:
    # Rough equivalent of git_config_open_default (not available in pygit2 yet)
    default_config_getters: _typing.ClassVar = [
        (GitConfigLevel.GLOBAL, GitConfig.get_global_config),   # ~/.gitconfig
        (GitConfigLevel.XDG,    GitConfig.get_xdg_config),      # ~/.config/git/config
        (GitConfigLevel.SYSTEM, GitConfig.get_system_config),   # /etc/gitconfig
        # Note: libgit2's git_config_open_default also adds GitConfigLevel.PROGRAMDATA for Windows,
        # but pygit2 doesn't expose git_config_find_programdata.
    ]
    assert (sorted([gcl for gcl, _ in default_config_getters], reverse=True)
            == [gcl for gcl, _ in default_config_getters]
            ), "default_config_getters should be sorted by decreasing globality"

    class SectionPatterns:
        SUBMODULE = _re.compile(r'^submodule "(.+)"$')
        REMOTE = _re.compile(r'^remote "(.+)"$')
        BRANCH = _re.compile(r'^branch "(.+)"$')

    @staticmethod
    def sanitize_key(key: str | tuple[str, ...]) -> str:
        if type(key) is tuple:
            return ".".join(key)
        assert type(key) is str
        return key

    @staticmethod
    def path_for_level(level: GitConfigLevel, missing_dir_ok=False):
        if not (GitConfigLevel.PROGRAMDATA <= level <= GitConfigLevel.GLOBAL):
            raise NotImplementedError(f"unsupported level {level}")

        # Several paths may be concatenated with GIT_PATH_LIST_SEPARATOR,
        # which git2/common.h defines as ":" (or ";" on Windows).
        # pygit2 doesn't expose this, but it appears to match os.pathsep.
        search_paths = _pygit2settings.search_path[level].split(_os.pathsep)
        search_paths = [path for path in search_paths if path]  # filter out any empty paths

        if not search_paths:
            raise KeyError(f"no search path is set up for git config level {level}")

        try:
            path = next(path for path in search_paths if _isdir(path))
        except StopIteration as exc:
            if not missing_dir_ok:
                raise FileNotFoundError(f"no parent dir exists for git config level {level}") from exc
            path = search_paths[0]

        # Select appropriate filename (see libgit2/config.h: GIT_CONFIG_FILENAME_SYSTEM, etc.)
        if level == GitConfigLevel.SYSTEM:
            filename = "gitconfig"
        elif level == GitConfigLevel.GLOBAL:
            filename = ".gitconfig"
        else:
            filename = "config"

        path = _joinpath(path, filename)
        return path

    @staticmethod
    def ensure_file(level: GitConfigLevel = GitConfigLevel.GLOBAL) -> GitConfig:
        try:
            if level == GitConfigLevel.GLOBAL:
                return GitConfig.get_global_config()
            elif level == GitConfigLevel.SYSTEM:
                return GitConfig.get_system_config()
            elif level == GitConfigLevel.XDG:
                return GitConfig.get_xdg_config()
            else:
                raise NotImplementedError("ensure_git_config_file: unsupported level")
        except OSError:
            # Last resort, create file
            path = GitConfigHelper.path_for_level(level, missing_dir_ok=True)
            _os.makedirs(_dirname(path), exist_ok=True)
            return GitConfig(path)

    @staticmethod
    def is_empty(config_path: str):
        with open(config_path, encoding="utf-8") as f:
            data = f.read()
        return not data or data.isspace()

    @staticmethod
    def delete_section(config_path: str, *section_key_tokens: str):
        config = GitConfig(config_path)

        assert len(section_key_tokens) == 2
        section_key = ".".join(section_key_tokens)
        section_prefix = section_key + "."

        to_delete = [entry.name for entry in config if entry.name.startswith(section_prefix)]

        for entry_name in to_delete:
            del config[entry_name]

        GitConfigHelper.scrub_empty_section(config_path, *section_key_tokens)

        return bool(to_delete)

    @staticmethod
    def scrub_empty_section(config_path: str, *section_key_tokens: str):
        num_tokens = len(section_key_tokens)
        if num_tokens == 1:
            section_key = section_key_tokens[0]
        elif num_tokens == 2:
            prefix, name = section_key_tokens
            name = name.replace('"', r'\"')
            section_key = f'{prefix} "{name}"'
        else:
            raise NotImplementedError("scrub_empty_section: Section key must be 1 or 2 tokens")

        # strict=False: Don't raise DuplicateSectionError.
        # If there are duplicate sections, their contents are merged.
        ini = _configparser.ConfigParser(strict=False)
        ini.read(config_path)

        # Get section. Bail if it doesn't appear in the file.
        try:
            section = ini[section_key]
            assert isinstance(section, _configparser.SectionProxy)
        except KeyError:
            _logger.debug(f"scrub_empty_section: Section [{section_key}] doesn't appear, no scrubbing needed")
            return

        # Bail if section isn't empty.
        if len(section) != 0:
            _logger.debug(f"scrub_empty_section: Section [{section_key}] isn't empty, won't scrub")
            return

        _logger.debug(f"scrub_empty_section: Scrubbing empty section [{section_key}]")

        # We could call ini.remove_section(section_key) then write the ini back
        # to disk, but this destroys the file's formatting. So, remove the
        # offending lines surgically. There may be multiple lines to remove if
        # the config file is fragmented (several remnants of empty sections).
        with open(config_path, encoding="utf-8") as f:
            lines = f.readlines()
        evict = f"[{section_key}]\n"
        lines = [line for line in lines if line != evict]

        # Write result to disk.
        timestamp = datetime.datetime.now(datetime.UTC).timestamp()

        temp_path = config_path + f".{timestamp}.new.tmp"
        backup_path = config_path + f".{timestamp}.old.tmp"

        with open(temp_path, "w", encoding="utf-8") as f:
            f.writelines(lines)

        _os.rename(config_path, backup_path)
        _os.rename(temp_path, config_path)
        _os.unlink(backup_path)

    @classmethod
    def global_identity(cls) -> tuple[str, str, GitConfigLevel]:
        """
        Returns the name and email set in the global `.gitconfig` file.
        If the global identity isn't set, this function returns blank strings.
        """

        name = ""
        email = ""
        level = GitConfigLevel.HIGHEST_LEVEL

        for getter_level, getter in cls.default_config_getters:
            try:
                config = getter()
            except OSError:
                # "The global file '.gitconfig' doesn't exist: No such file or directory"
                continue

            with _suppress(KeyError):
                name = config["user.name"]
                level = getter_level

            with _suppress(KeyError):
                email = config["user.email"]
                level = getter_level

            if name and email:
                break

        return name, email, level

    @classmethod
    def get_default_value(cls, key: str | tuple) -> str:
        key = cls.sanitize_key(key)
        for _dummy, getter in cls.default_config_getters:
            with _suppress(OSError, KeyError):
                config = getter()
                return config[key]
        return ""

    @staticmethod
    def unescape(s: str):
        return s.encode('raw_unicode_escape').decode('unicode_escape')

    @staticmethod
    def walk_sections(config: _configparser.ConfigParser, section_pattern: _re.Pattern):
        for section_name in config.sections():
            match = section_pattern.fullmatch(section_name)
            if not match:
                continue

            section = config[section_name]
            name = GitConfigHelper.unescape(match.group(1))
            yield name, section


class Repo(_VanillaRepository):
    """
    Drop-in replacement for pygit2.Repository with convenient front-ends to common git operations.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._cached_config_stats = {}
        self._cached_config_stats_for_key = {}
        self._cached_config_parsers = {}
        self._cached_remote_list = []
        self._cached_upstream_table = {}
        self._cached_submodule_table = {}
        self._cached_commondir = ""

    def __del__(self):
        _logger.debug("__del__ Repo")
        self.free()

    @property
    def head_tree(self) -> Tree:
        return self.head.peel(Tree)

    @property
    def head_commit(self) -> Commit:
        return self.head.peel(Commit)

    @property
    def head_commit_id(self) -> Oid:
        return self.head_commit.id

    @property
    def head_commit_message(self) -> str:
        return self.head_commit.message

    @property
    def head_branch_shorthand(self) -> str:
        return self.head.shorthand

    @property
    def head_branch_fullname(self) -> str:
        return self.head.name

    def peel_commit(self, commit_id: Oid | str) -> Commit:
        return self[commit_id].peel(Commit)

    def peel_blob(self, blob_id: Oid | str) -> Blob:
        return self[blob_id].peel(Blob)

    def peel_tree(self, tree_id: Oid | str) -> Tree:
        return self[tree_id].peel(Tree)

    def in_workdir(self, path: str) -> str:
        """Return an absolutized version of `path` within this repo's workdir."""
        assert not _isabs(path)
        p = _joinpath(self.workdir, path)
        if not p.startswith(self.workdir):
            raise ValueError("Won't create absolute path outside workdir")
        return p

    def is_in_workdir(self, path: str, include_gitdir: bool = False) -> bool:
        path_obj = _Path(path)
        return (path_obj.is_relative_to(self.workdir)
                and (include_gitdir or not path_obj.is_relative_to(self.path)))

    @property
    def commondir(self) -> str:
        """
        Return the absolute path to this worktree's $GIT_COMMON_DIR.
        If this Repo isn't a worktree, the return value resolves to the same
        path as `Repo.path`.
        """
        if not self._cached_commondir:
            path = self.path
            commondir = _Path(path, "commondir")
            if commondir.exists():
                path = commondir.read_text().rstrip()
                path_obj = commondir.parent / path
                path_obj = path_obj.resolve(strict=True)
                path = str(path_obj)
                # For compatibility with Repository.path, tack a '/' to the end
                assert not path.endswith(_os.sep)
                path += _os.sep
            self._cached_commondir = path
        return self._cached_commondir

    def in_gitdir(self, path: str, common: bool = True) -> str:
        """
        Return an absolutized version of `path` within this repo's .git directory.

        If `common` is False, return `path` inside the worktree's private $GIT_DIR.
        Otherwise, return `path` inside $GIT_COMMON_DIR, i.e. the ".git" directory
        common to all worktrees. (See manpages for gitrepository-layout git-worktree)

        The `common` argument has no effect if this repo isn't a worktree.
        """
        assert not _isabs(path)

        parent = self.commondir if common else self.path
        parent_resolved = _Path(parent).resolve()
        p = _Path(parent, path).resolve()

        if not p.is_relative_to(parent_resolved):
            raise ValueError("Won't resolve absolute path outside gitdir")

        return str(p)

    def refresh_index(self, force: bool = False):
        """
        Reload the index. Call this before manipulating the staging area
        to ensure any external modifications are taken into account.

        This is a fairly cheap operation if the index hasn't changed on disk,
        unless you pass force=True.
        """
        self.index.read(force)

    def get_unstaged_changes(self, update_index: bool = False, show_binary: bool = False, context_lines: int = 3) -> Diff:
        """
        Get a Diff of unstaged changes in the working directory.

        In other words, this function compares the workdir to the index.
        """

        flags = (DiffOption.INCLUDE_UNTRACKED
                 | DiffOption.RECURSE_UNTRACKED_DIRS
                 | DiffOption.SHOW_UNTRACKED_CONTENT
                 | DiffOption.INCLUDE_TYPECHANGE
                 )

        # Don't attempt to update the index if the repo is locked for writing,
        # or the index is locked by another program
        update_index = (update_index
                        and _os.access(self.path, _os.W_OK)
                        and not _exists(_joinpath(self.path, "index.lock")))

        # UPDATE_INDEX may improve performance for subsequent diffs if the
        # index was stale, but this requires the repo to be writable.
        if update_index:
            _logger.debug("UPDATE_INDEX")
            flags |= DiffOption.UPDATE_INDEX

        if show_binary:
            flags |= DiffOption.SHOW_BINARY

        dirty_diff = self.diff(None, None, flags=flags, context_lines=context_lines)
        # dirty_diff.find_similar()  #-- it seems that find_similar cannot find renames in unstaged changes, so don't bother
        return dirty_diff

    def get_staged_changes(self, fast: bool = False, show_binary: bool = False, context_lines: int = 3) -> Diff:
        """
        Get a Diff of the staged changes.

        In other words, this function compares the index to HEAD.
        """

        flags = DiffOption.INCLUDE_TYPECHANGE

        if show_binary:
            flags |= DiffOption.SHOW_BINARY

        if self.head_is_unborn:  # can't compare against HEAD (empty repo or branch pointing nowhere)
            index_tree_id = self.index.write_tree()
            tree = self.peel_tree(index_tree_id)
            return tree.diff_to_tree(swap=True, flags=flags, context_lines=context_lines)
        else:
            # compare HEAD to index
            stage_diff: Diff = self.diff('HEAD', None, cached=True, flags=flags, context_lines=context_lines)
            if not fast:
                stage_diff.find_similar()
            return stage_diff

    @property
    def any_conflicts(self) -> bool:
        """True if there are any conflicts in the index."""
        return bool(self.index.conflicts)

    @property
    def any_staged_changes(self) -> bool:
        """True if there are any staged changes in the index."""
        return 0 != len(self.get_staged_changes(fast=True, context_lines=0))
        # ---This also works, but it's slower.
        # status = repo.status(untracked_files="no")
        # return any(0 != (flag & FileStatus.INDEX_MASK) for flag in status.values())

    def checkout_local_branch(self, name: str):
        """Switch to a local branch."""
        branch = self.branches.local[name]
        refname = branch.name
        self.checkout_ref(refname)

    def checkout_ref(self, refname: str):
        """Enter detached HEAD on the commit pointed to by a ref."""
        with CheckoutBreakdown() as callbacks:
            self.checkout(refname, callbacks=callbacks)

    def checkout_commit(self, commit_id: Oid):
        """Enter detached HEAD on a commit."""
        commit = self.peel_commit(commit_id)
        with CheckoutBreakdown() as callbacks:
            self.checkout_tree(commit.tree, callbacks=callbacks)
            self.set_head(commit_id)

    def rename_local_branch(self, name: str, new_name: str) -> Branch:
        """Rename a local branch."""
        # TODO: if the branch tracks an upstream branch, issue a warning that it won't be renamed on the server
        branch = self.branches.local[name]
        branch.rename(new_name)
        self.scrub_empty_config_section("branch", name)
        return branch

    def delete_local_branch(self, name: str):
        """Delete a local branch."""
        # TODO: if remote-tracking, let user delete upstream too?
        self.branches.local.delete(name)
        self.scrub_empty_config_section("branch", name)

    def scrub_empty_config_section(self, *section_key_tokens: str):
        """
        libgit2 leaves behind empty sections in the config file after deleting
        or renaming a remote or branch. Over time, this litters the config
        file with empty entries, which slows down operations that parse the
        config. So, use this function to nip that in the bud.
        """
        config_path = _joinpath(self.path, "config")
        GitConfigHelper.scrub_empty_section(config_path, *section_key_tokens)

    def create_branch_on_head(self, name: str) -> Branch:
        """Create a local branch pointing to the commit at the current HEAD."""
        return self.create_branch(name, self.head_commit)

    def create_branch_from_commit(self, name: str, commit_id: Oid) -> Branch:
        """Create a local branch pointing to the given commit id."""
        commit = self.peel_commit(commit_id)
        branch = self.create_branch(name, commit)
        return branch

    def listall_remote_branches(self, value_style: _typing.Literal["strip", "shorthand", "refname"] = "strip") -> dict[str, list[str]]:
        names: dict[str, list[str]] = {}

        # Create empty lists for all remotes (including branchless remotes)
        for remote in self.remotes:
            names[remote.name] = []

        for ref in self.references.iterator():
            prefix, shorthand = RefPrefix.split(ref.name)

            if prefix != RefPrefix.REMOTES:
                continue

            if ref.name.endswith("/HEAD"):
                # Skip refs/remotes/*/HEAD (the remote's default branch).
                # The ref file (.git/refs/remotes/*/HEAD) is created ONCE when first cloning the repository,
                # and it's never updated again automatically, even if the default branch has changed on the remote.
                # It's a symbolic branch, so looking up a stale version of the remote's HEAD may raise KeyError.
                # It's just not worth the trouble.
                # See: https://stackoverflow.com/questions/8839958
                continue

            # Skip symbolic refs (e.g., repo tool's refs/remotes/m/master -> origin/master).
            # We only show direct refs in real remotes; resolving would duplicate the target
            # (e.g. origin/master would appear twice in the upstream context menu).
            if ref.type == ReferenceType.SYMBOLIC:
                continue

            remote_name, branch_name = split_remote_branch_shorthand(shorthand)
            # Skip references without a branch name (e.g., refs/remotes/git-svn from git svn clone)
            if not branch_name:
                continue
            # Skip references that don't match any known remote (e.g., stale refs from deleted remotes)
            if remote_name not in names:
                continue
            if value_style == "strip":
                value = branch_name
            elif value_style == "shorthand":
                value = shorthand
            elif value_style == "refname":
                value = ref.name
            else:
                raise NotImplementedError(f"unsupported value_style {value_style}")
            names[remote_name].append(value)

        return names

    def listall_tags(self) -> list[str]:
        return [
            name.removeprefix(RefPrefix.TAGS)
            for name in self.listall_references()
            if name.startswith(RefPrefix.TAGS)
        ]

    def edit_upstream_branch(self, local_branch_name: str, remote_branch_name: str):
        local_branch = self.branches.local[local_branch_name]
        if remote_branch_name:
            remote_branch = self.branches.remote[remote_branch_name]
            local_branch.upstream = remote_branch
        else:
            if local_branch.upstream is not None:
                local_branch.upstream = None

    def get_remote_skipfetchall(self, remote_name: str) -> bool:
        key = GitConfigHelper.sanitize_key(("remote", remote_name, "skipFetchAll"))
        try:
            return self.config.get_bool(key)
        except KeyError:
            return False

    def set_remote_skipfetchall(self, remote_name: str, value: bool):
        key = GitConfigHelper.sanitize_key(("remote", remote_name, "skipFetchAll"))

        # Set skipFetchAll only if it's True, otherwise delete it
        if value:
            self.config[key] = value
        else:
            with _suppress(KeyError):
                del self.config[key]

    def get_commit_message(self, commit_id: Oid) -> str:
        commit = self.peel_commit(commit_id)
        return commit.message

    def create_commit_on_head(
            self,
            message: str,
            author: Signature | None = None,
            committer: Signature | None = None
    ) -> Oid:
        """
        Create a commit with the contents of the index tree.
        Use the commit at HEAD as the new commit's parent.
        If `author` or `committer` are not overridden, use the repository's default_signature.
        """

        # Prep ref to update
        if self.head_is_detached:
            ref_to_update = "HEAD"
        else:
            # Get the ref name pointed to by HEAD, but DON'T use repo.head! It won't work if HEAD is unborn.
            # Both git and libgit2 store a default branch name in .git/HEAD when they init a repo,
            # so we should always have a ref name, even though it might not point to anything.
            target = self.lookup_reference("HEAD").target
            assert isinstance(target, str), "HEAD isn't a symbolic reference!"
            ref_to_update = target

        # Prep parent list
        if self.head_is_unborn:
            parents = []
        else:
            # Always take HEAD commit as 1st parent
            parents = [self.head_commit_id]

        # If a merge was in progress, add merge heads to parent list
        parents += self.listall_mergeheads()
        assert len(set(parents)) == len(parents), "duplicate heads!"

        # Get id of index tree
        index_tree_id = self.index.write_tree()

        # Take default signature now to prevent any timestamp diff between
        # author and committer. Note that there may not be a fallback signature
        # if the system's git config isn't set up.
        try:
            fallback_signature = self.default_signature
        except KeyError:
            fallback_signature = None
            assert author is not None, "fallback signature missing - author signature must be provided"
            assert committer is not None, "fallback signature missing - committer signature must be provided"

        # Create the commit
        new_commit_id = self.create_commit(
            ref_to_update,
            author or fallback_signature,
            committer or fallback_signature,
            message,
            index_tree_id,
            parents
        )

        # Clear repository state to conclude a merge/cherrypick operation
        self.state_cleanup()

        # Repository.create_commit flushes the staged changes from the in-memory index.
        # Write the index to disk so that other applications can pick up the updated staging area.
        self.index.write()

        assert not self.head_is_unborn, "HEAD is still unborn after we have committed!"

        return new_commit_id

    def amend_commit_on_head(
            self,
            message: str,
            author: Signature | None = None,
            committer: Signature | None = None
    ) -> Oid:
        """
        Amend the commit at HEAD with the contents of the index tree.
        If `author` is None, don't replace the original commit's author.
        If `committer` is None, use default_signature for the committer's signature.
        """
        index_tree_id = self.index.write_tree(self)
        new_commit_id = self.amend_commit(
            self.head_commit,
            'HEAD',
            message=message,
            author=author,
            committer=committer or self.default_signature,
            tree=index_tree_id
        )
        return new_commit_id

    def commit_id_from_refname(self, refname: str) -> Oid:
        reference = self.references[refname]
        commit: Commit = reference.peel(Commit)
        return commit.id

    def commit_id_from_tag_name(self, tagname: str) -> Oid:
        assert not tagname.startswith("refs/")
        return self.commit_id_from_refname(RefPrefix.TAGS + tagname)

    def map_refs_to_ids(self, include_stashes=False) -> dict[str, Oid]:
        """
        Return commit oids at the tip of all branches, tags, etc. in the repository.

        To ensure a consistent outcome across multiple walks of the same commit graph,
        the oids are sorted by ascending commit time.
        """

        tips: list[tuple[str, Commit]] = []

        supported_prefixes = (RefPrefix.HEADS, RefPrefix.REMOTES, RefPrefix.TAGS)

        for ref in self.listall_reference_objects():
            name: str = ref.name

            # Any text file containing a hash will be picked up by libgit2 if it is in refs/.
            # Exclude those if they don't match any of our supported ref prefixes.
            # Note that this also excludes refs/stash - we deal with stashes separately.
            if not name.startswith(supported_prefixes):
                continue

            # Skip symbolic references
            if ref.type != ReferenceType.DIRECT:
                continue

            # Skip non-commit references
            obj = ref.peel()
            if obj.type != ObjectType.COMMIT:
                continue

            assert isinstance(obj, Commit)
            tips.append((name, obj))

        if include_stashes:
            for i, stash in enumerate(self.listall_stashes()):
                try:
                    commit = self.peel_commit(stash.commit_id)
                    tips.append((f"stash@{{{i}}}", commit))
                except InvalidSpecError as e:
                    _logger.info(f"{e} - Skipping stash '{stash.message}'")

        # Always add 'HEAD' if we have one.
        # Do so *just* before reinserting the tips in chronological order below.
        # This will sort the checked-out branch at the *end* of the list
        # if there is another tip that shares the exact same timestamp.
        # (Refs pushed into Walker are Last-In First-Out, so the local branch will
        # come out on top e.g. when pushing a branch to a remote, then amending
        # the top commit with the same author/committer signatures.)
        try:
            tips.append(("HEAD", self.head_commit))
        except (GitError, InvalidSpecError):
            pass  # Skip detached/unborn head

        # Reinsert all tips in chronological order
        # (In Python 3.7+, dict key order is stable)
        tips.sort(key=lambda item: item[1].commit_time)
        return {ref: commit.id for ref, commit in tips}

    def listall_refs_pointing_at(self, commit_id: Oid, include_stashes=False):
        refs = []

        # Detached HEAD isn't in repo.references
        if self.head_is_detached and type(self.head.target) is Oid and self.head.target == commit_id:
            refs.append('HEAD')

        for ref in self.references.objects:
            ref_key = ref.name

            if type(ref.target) is not Oid:
                # Symbolic reference
                _logger.debug(f"Skipping symbolic reference {ref_key} --> {ref.target}")
                continue

            if ref.target != commit_id:
                continue

            assert ref_key.startswith("refs/")

            if ref_key == "refs/stash":
                # This is just the topmost stash.
                # Exhaustive stash info must be obtained separately.
                continue

            refs.append(ref_key)

        if include_stashes:
            for stash_index, stash in enumerate(self.listall_stashes()):
                if stash.commit_id == commit_id:
                    refs.append(f"stash@{{{stash_index}}}")

        return refs

    def stage_files(self, patches: list[Patch]):
        index = self.index
        error = MultiFileError()

        for patch in patches:
            path = patch.delta.new_file.path
            if patch.delta.new_file.mode == FileMode.TREE and path.endswith("/"):
                path = path.removesuffix("/")
            try:
                if patch.delta.status == DeltaStatus.DELETED:
                    index.remove(path)
                else:
                    index.add(path)
            except GitError as gitError:
                error.add_file_error(path, gitError)
            else:
                error.add_file_success()

        index.write()

        if error:
            raise error

    def discard_mode_changes(self, paths: list[str]):
        """
        Discards mode changes in the given files.
        """

        self.refresh_index()  # in case an external program modified the staging area
        index = self.index

        # Reset files to their mode in the staged (index) tree
        for p in paths:
            try:
                mode = index[p].mode
            except KeyError:
                continue
            if mode in [FileMode.BLOB, FileMode.BLOB_EXECUTABLE]:
                _os.chmod(self.in_workdir(p), mode)

    def create_stash(self, message: str, paths: list[str]) -> Oid:
        """
        Creates a stash that backs up all changes to the given files.
        Does NOT remove the changes from the workdir!
        (You can call 'git clean'/'git restore' manually afterwards).
        """

        assert paths, "path list cannot be empty"

        try:
            signature = self.default_signature
        except (KeyError, ValueError):
            # Allow creating a stash if the identity isn't set
            signature = Signature(name="UNKNOWN", email="UNKNOWN")

        commit_id = self.stash(
            stasher=signature,
            message=message,
            keep_index=False,
            keep_all=True,
            include_untracked=False,
            include_ignored=False,
            paths=paths)

        return commit_id

    def find_stash_index(self, commit_id: Oid) -> int:
        """
        Libgit2 takes an index number to apply/pop/drop stashes. However, it's
        unsafe to cache such an index for the GUI. Instead, we cache the commit ID
        of the stash, and we only convert that to an index when we need to perform
        an operation on the stash. This way, we'll always manipulate the stash
        intended by the user, even if the indices change outside our control.
        """
        try:
            return next(i for i, stash in enumerate(self.listall_stashes())
                        if stash.commit_id == commit_id)
        except StopIteration as exc:
            raise KeyError(f"Stash not found: {commit_id}") from exc

    def applies_breakdown(self, patch_data: bytes | str, location=ApplyLocation.WORKDIR) -> Diff:
        diff = Diff.parse_diff(patch_data)
        error = MultiFileError()

        # Attempt to apply every patch in the diff separately, so we can report which file an error pertains to
        for patch in diff:
            patch_data = patch.data

            # Work around libgit2 bug: If a patch lacks the "index" line (as our partial patches do),
            # then libgit2 fails to recreate the "---" and "+++" lines. Then it can't parse its own output.
            patch_lines = patch_data.splitlines(True)
            if len(patch_lines) >= 2 and patch_lines[0].startswith(b"diff --git ") and patch_lines[1].startswith(b"@@"):
                header = patch_lines[0].strip().decode("utf-8")
                match = _re.match(DIFF_HEADER_PATTERN, header)
                if match:
                    patch_lines.insert(1, ("--- " + match[1] + "\n").encode("utf-8"))
                    patch_lines.insert(2, ("+++ " + match[2] + "\n").encode("utf-8"))
                    patch_data = b"".join(patch_lines)

            patch_diff = Diff.parse_diff(patch_data)  # can we extract a diff from the patch without re-parsing it?
            try:
                self.applies(patch_diff, location, raise_error=True)
            except (GitError, OSError) as exc:
                error.add_file_error(patch.delta.old_file.path, exc)
            else:
                error.add_file_success()

        if error:
            raise error

        return diff

    # Override pygit2.Repository.apply()
    def apply(self,
              patch_data_or_diff: bytes | str | Diff,
              location: ApplyLocation = ApplyLocation.WORKDIR
              ):
        if isinstance(patch_data_or_diff, bytes | str):
            diff = Diff.parse_diff(patch_data_or_diff)
        elif isinstance(patch_data_or_diff, Diff):
            diff = patch_data_or_diff
        else:
            raise TypeError("patch_data_or_diff must be bytes, str, or Diff")

        super().apply(diff, location)

    def get_submodule_workdir(self, submo_key: str) -> str:
        submo = self.submodules[submo_key]
        return self.in_workdir(submo.path)

    def submodule_dotgit_present(self, submo_path: str) -> bool:
        path = self.in_workdir(submo_path)
        path = _joinpath(path, ".git")
        return _exists(path)

    def fast_forward_branch(self, local_branch_name: str, target_branch_name: str = ""):
        """
        Fast-forward a local branch to another branch (local or remote).
        Return True if the local branch was up to date.
        Raise DivergentBranchesError if fast-forwarding is impossible.

        If target_branch_name is omitted, attempt to fast-forward to the
        branch's upstream (will not fetch it beforehand).
        """
        assert not local_branch_name.startswith("refs/")
        lb = self.branches.local[local_branch_name]

        if not target_branch_name:
            rb = lb.upstream
            if not rb:
                raise ValueError("Local branch does not track a remote branch")
        else:
            assert isinstance(target_branch_name, str)
            target_prefix, target_shorthand = RefPrefix.split(target_branch_name)
            if target_prefix == RefPrefix.REMOTES:
                rb = self.branches.remote[target_shorthand]
            elif target_prefix == RefPrefix.HEADS:
                rb = self.branches.local[target_shorthand]
            else:
                rb = self.branches[target_shorthand]

        merge_analysis, merge_pref = self.merge_analysis(rb.target, RefPrefix.HEADS + local_branch_name)
        _logger.debug(f"Merge analysis: {merge_analysis!r}. Merge preference: {merge_pref!r}.")

        if merge_analysis & MergeAnalysis.UP_TO_DATE:
            # Local branch is up to date with remote branch, nothing to do.
            return True

        elif merge_analysis == (MergeAnalysis.NORMAL | MergeAnalysis.FASTFORWARD):
            # Go ahead and fast-forward.

            # First, we need to check out the tree pointed to by the remote branch. This step is necessary,
            # otherwise the contents of the commits we're pulling will spill into the unstaged area.
            # Note: checkout_tree defaults to a safe checkout, so it'll raise GitError if any uncommitted changes
            # affect any of the files that are involved in the pull.
            if lb.is_checked_out():
                with CheckoutBreakdown() as callbacks:
                    self.checkout_tree(rb.peel(Tree), callbacks=callbacks)

            # Then make the local branch point to the same commit as the remote branch.
            lb.set_target(rb.target)

        elif merge_analysis == MergeAnalysis.NORMAL:
            # Can't FF. Divergent branches?
            raise DivergentBranchesError(lb, rb)

        else:
            # Unborn or something...
            raise NotImplementedError(f"Cannot fast-forward with {merge_analysis!r}.")

    def repo_name(self):
        return _basename(_normpath(self.workdir))

    def get_local_identity(self) -> tuple[str, str]:
        """
        Return the name and email set in the repository's local config.
        If an identity isn't set specifically for this repo, this function returns blank strings.
        """

        config = self.config_snapshot
        local = {}
        for key in ["user.name", "user.email"]:
            with _suppress(KeyError):
                entry = config._get_entry(key)
                if GitConfigLevel.LOCAL <= entry.level:  # <= GitConfigLevel.WORKTREE: -----requires pygit2 1.15
                    local[key] = entry.value

        name = local.get("user.name", "")
        email = local.get("user.email", "")
        return name, email

    def has_local_identity(self):
        name, email = self.get_local_identity()
        return bool(name or email)

    def delete_tag(self, tagname: str):
        assert not tagname.startswith("refs/")
        refname = RefPrefix.TAGS + tagname
        assert refname in self.references
        self.references.delete(refname)

    def get_config_value(self, key: str | tuple):
        key = GitConfigHelper.sanitize_key(key)
        try:
            return self.config[key]
        except KeyError:
            return ""

    def set_config_value(self, key: str | tuple, value: str):
        key = GitConfigHelper.sanitize_key(key)
        if value:
            self.config[key] = value
        else:
            with _suppress(KeyError):
                del self.config[key]

    def get_reset_merge_file_list(self):
        self.index.read()

        staged_diff = self.diff('HEAD', None, cached=True)  # staged - index to head
        unstaged_diff = self.diff(None, None)  # unstaged - index to workdir

        staged_paths = [p.delta.new_file.path for p in staged_diff]
        unstaged_paths = [p.delta.new_file.path for p in unstaged_diff if p.delta.status != DeltaStatus.CONFLICTED]

        intersection = set(staged_paths).intersection(unstaged_paths)

        if intersection:
            error = MultiFileError()
            for path in intersection:
                error.add_file_error(path, None)
            raise error

        return staged_paths

    @property
    def message_without_conflict_comments(self) -> str:
        message = self.message
        clean_message = ""
        for line in message.splitlines(keepends=True):
            if not line.startswith('#'):
                clean_message += line
        return clean_message.strip()

    def set_message(self, message: str):
        with open(_joinpath(self.path, "MERGE_MSG"), "w") as mergemsg:
            mergemsg.write(message)

    # -------------------------------------------------------------------------
    # Config parsers ("listall" APIs)

    def listall_remotes_fast(self):
        """
        Faster alternative to `list(repo.remotes.names())`.
        """
        config_path = self.in_gitdir("config", common=True)
        try:
            # strict=False to bypass DuplicateSectionError
            config = self._get_cached_config("listall_remotes_fast", config_path, strict=False)
        except OSError:
            self._cached_remote_list = []
        except _UnchangedConfigFile:
            pass
        else:
            self._cached_remote_list = Repo._listall_remotes_from_config(config)

        return self._cached_remote_list

    def listall_upstreams_fast(self) -> dict[str, str]:
        """
        Much faster alternative to looking up `Branch.upstream_name` in every branch.
        """
        config_path = self.in_gitdir("config", common=True)
        try:
            # strict=False to bypass DuplicateSectionError
            config = self._get_cached_config("listall_upstreams", config_path, strict=False)
        except OSError:
            self._cached_upstream_table = {}
        except _UnchangedConfigFile:
            pass
        else:
            self._cached_upstream_table = Repo._listall_upstreams_from_config(config)

        return self._cached_upstream_table

    def listall_submodules(self):  # pragma: no cover
        """
        Don't use listall_submodules, it's slow and it omits critical
        information about submodule names.
        Note that if a submodule's name is different from its path,
        SubmoduleCollection[path].name will NOT return the correct name!
        """
        warnings.warn("Don't use this", DeprecationWarning)
        raise DeprecationWarning("Don't use listall_submodules")

    def listall_submodules_fast(self) -> list[str]:
        """
        Faster drop-in replacement for pygit2's Repository.listall_submodules (which can be very slow).
        Return a list of submodule workdir paths within the root repo's workdir.
        """
        return list(self.listall_submodules_dict().values())

    def listall_submodules_dict(self, config_text="") -> dict[str, str]:
        """
        Return a dict of submodule names to paths.
        Information is sourced from .gitmodules (or config_text if given).

        You should use this instead of pygit2's Repository.listall_submodules,
        because the latter doesn't give you information about submodule names.
        It is also faster.
        """
        # We use ConfigParser instead of pygit2.Config so that we can load the
        # config from a blob in memory.

        if config_text:
            config = _configparser.ConfigParser()
            config.read_string(config_text)
            return Repo._listall_submodules_from_config(config)

        config_path = self.in_workdir(DOT_GITMODULES)
        try:
            config = self._get_cached_config("listall_submodules_dict", config_path, strict=True)
        except OSError:
            self._cached_submodule_table = {}
        except _UnchangedConfigFile:
            pass
        else:
            self._cached_submodule_table = Repo._listall_submodules_from_config(config)
        return self._cached_submodule_table

    def listall_submodules_dict_at_head(self) -> dict[str, str]:
        try:
            gitmodules_blob: Blob = self.head_tree[".gitmodules"].peel(Blob)
        except (KeyError,  # .gitmodules doesn't exist
                ValueError):  # .gitmodules not Blob (probably Tree)
            return {}
        old_gitmodules = gitmodules_blob.data.decode("utf-8")
        return self.listall_submodules_dict(config_text=old_gitmodules)

    def _get_cached_config(self, cache_key: str, path: str, strict: bool) -> _configparser.ConfigParser:
        new_stat = _os.stat(path)

        with _suppress(KeyError):
            key_stat = self._cached_config_stats_for_key[cache_key]
            if key_stat == new_stat:
                raise _UnchangedConfigFile()

        self._cached_config_stats_for_key[cache_key] = new_stat

        with _suppress(KeyError):
            known_stat = self._cached_config_stats[path]
            if new_stat == known_stat:
                return self._cached_config_parsers[path]

        config = _configparser.ConfigParser(strict=strict)
        config.read(path)
        self._cached_config_stats[path] = new_stat
        self._cached_config_parsers[path] = config
        return config

    @staticmethod
    def _listall_remotes_from_config(config: _configparser.ConfigParser):
        remotes = []

        for remote_name, section in GitConfigHelper.walk_sections(config, GitConfigHelper.SectionPatterns.REMOTE):
            if not section.items():  # Skip empty sections
                continue
            remotes.append(remote_name)

        return remotes

    @staticmethod
    def _listall_upstreams_from_config(config: _configparser.ConfigParser):
        upstreams = {}

        for branch_name, section in GitConfigHelper.walk_sections(config, GitConfigHelper.SectionPatterns.BRANCH):
            try:
                remote_name = section["remote"]
                remote_ref = section["merge"]
            except KeyError:
                continue

            remote_ref_prefix, remote_branch = RefPrefix.split(remote_ref)
            if remote_ref_prefix != RefPrefix.HEADS:
                _logger.warning(f"unexpected prefix in config value {section}.merge: '{remote_branch}'")
                continue

            remote_branch = GitConfigHelper.unescape(remote_branch)
            remote_name = GitConfigHelper.unescape(remote_name)

            upstreams[branch_name] = f"{remote_name}/{remote_branch}"

        return upstreams

    @staticmethod
    def _listall_submodules_from_config(config: _configparser.ConfigParser):
        submodules = {}

        for submodule_name, section in GitConfigHelper.walk_sections(config, GitConfigHelper.SectionPatterns.SUBMODULE):
            try:
                path = section["path"]
            except KeyError:
                continue

            path = GitConfigHelper.unescape(path)
            submodules[submodule_name] = path

        return submodules

    def apply_filters_to_workdir(self, path: str) -> bytes:
        # See if this file is subject to any filters
        try:
            filters = self.load_filter_list(path)
        except AttributeError:  # pragma: no cover (old pygit2 compatibility)
            # TODO: Remove this once we can stop supporting pygit2 <= 1.19.1.
            #       This applies filters, but it also writes the resulting blob
            #       to the ODB on disk, which isn't optimal for transient contents
            #       that may change frequently (unstaged files).
            oid = self.create_blob_fromworkdir(path)
            blob = self.peel_blob(oid)
            return blob.data

        if filters is None:
            # No filters, read file from disk
            return _Path(self.in_workdir(path)).read_bytes()
        else:
            # Apply filters in a buffer
            return filters.apply_to_file(self, path)


class RepoContext:
    def __init__(self, repo_or_path: Repo | str | _Path, flags=RepositoryOpenFlag.DEFAULT, write_index=False):
        assert isinstance(repo_or_path, Repo | str | _Path)
        if isinstance(repo_or_path, Repo):
            self.repo = repo_or_path
            assert flags == 0, "flags are ignored if passing in a Repo"
        else:
            path = repo_or_path
            self.repo = Repo(path, flags)
        self.write_index = write_index

    def __enter__(self) -> Repo:
        return self.repo

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.write_index:
            self.repo.index.write()
        # repo.free() is necessary for correct test teardown on Windows
        self.repo.free()
        self.repo = None
